"""Scraper for North Carolina Supreme Court
CourtID: nc
Court Short Name: N.C.
Reviewer:
History:
    2014-05-01: Created by Brian Carver
    2014-08-04: Rewritten by Jon Andersen with complete backscraper
    2025-04-22: grossir, Update to OpinionSiteLinear
"""

import re
from datetime import datetime

from lxml import html

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    start_year = 1998
    current_year = datetime.today().year
    court = "sc"
    court_number = 1
    row_xpath = "//span[span[@class='title']] | //td[span[@class='title']]"
    date_xpath = "../../preceding-sibling::tr//strong[contains(text()[1], 'Filed:')] | ../preceding-sibling::tr/td[contains(text(), 'Rule 30e')]"
    date_regex = r"Filed: (?P<date>[\d\w ]+)"
    coa_rgx = re.compile(
        r"(?P<date>\d[\d \w]+)[\t\xa0\n]+- Rule 30e", flags=re.M
    )
    pattern = r"(?P<name>.*?)(?:[\s,]+\s*(?P<cite>\d+\s+NC(?:\s+App)?\s+\d+))?\s*\(\s*(?P<docket>[^\s]+)\s*-\s*(?P<status>[^)]+)\s*\)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.base_url = (
            "https://appellate.nccourts.org/opinion-filings/?c={}&year={}"
        )
        self.url = self.base_url.format(self.court, self.current_year)
        self.make_backscrape_iterable(kwargs)

    def _process_html(self):
        """Process the html

        :return: None
        """
        for row in self.html.xpath(self.row_xpath):
            title = row.xpath("string(span[@class='title'])")
            links = row.xpath("span[@class='title']/@onclick")
            summary = row.xpath("string(span[@class='desc'])")
            if not links:
                logger.warning("No link for row %s", title)
                continue
            url = links[0][13:-2].replace("http:", "https:")
            m = re.search(self.pattern, title)
            name, citation, docket, status = m.groups()
            author = row.xpath("string(span[@class='author']/i)").strip()
            per_curiam = author == "Per Curiam"
            date = self.extract_date(row)
            self.cases.append(
                {
                    "per_curiam": per_curiam,
                    "author": author if not per_curiam else "",
                    "docket": docket,
                    "status": status,
                    "name": name,
                    "url": url,
                    "date": date,
                    "summary": summary,
                    "citation": citation or "",
                }
            )

    def extract_date(self, row: html.HtmlElement) -> str:
        """Extract date str for opinion in row

        :param row: row to process
        :return: date string
        """
        date_block = row.xpath(self.date_xpath)[-1].text_content()
        if match := re.search(self.date_regex, date_block):
            return match.group("date")
        # Second pattern for ncctapp unpublished opinions
        return self.coa_rgx.search(date_block).group("date")

    def _download_backwards(self, year: int) -> None:
        """Build year URL and scrape

        :param year: year to scrape
        :return None
        """
        self.url = self.base_url.format(self.court, year)
        self.html = self._download()
        self._process_html()

    def make_backscrape_iterable(self, kwargs: dict) -> None:
        """Checks if backscrape start and end arguments have been passed
        by caller, and parses them accordingly

        :param kwargs: passed when initializing the scraper, may or
            may not contain backscrape controlling arguments

        :return None
        """
        start = kwargs.get("backscrape_start")
        end = kwargs.get("backscrape_end")

        start = int(start) if start else self.start_year
        end = int(end) + 1 if end else self.current_year

        self.back_scrape_iterable = range(start, end)



# Scraper for California's First District Court of Appeal
# CourtID: calctapp_1st
# Court Short Name: Cal. Ct. App.

import re

from juriscraper.opinions.united_states.state import cal


class Site(cal.Site):
    court_code = "A"
    division = "1st App. Dist."

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """

        pattern = re.compile(
            r"""
            (?:
            Trial\s+Court:\s+|
             (?:APPEALS?|Appeals?)\s+from\s+a\s+(?:judgments?|postjudgments?)(?:\s+order)?\s+of\s+the\s+|
             (?:APPEALS?|Appeals?)\s+from\s+(?:an\s+)?orders?\s+of\s+the\s+|
             (?:APPEALS?|Appeals?)\s+from\s+the\s+
             )
            (?P<lower_court>[A-Za-z\s]*?)(?=,|\n\s*\n|\.|Nos?.)
            """,
            re.X | re.DOTALL,
        )

        fall_back_pattern = re.compile(
            r"(?P<lower_court>[A-Za-z\s]+County\s+Super\. Ct\.)", re.MULTILINE
        )

        lower_court = ""
        if (match := pattern.search(scraped_text)) or (
            match := fall_back_pattern.search(scraped_text)
        ):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()

        if lower_court:
            return {
                "Docket": {
                    "appeal_from_str": lower_court,
                }
            }
        return {}



"""
Scraper for the Supreme Court of Tennessee, Workers Compensation Panel
CourtID: tenn
Court Short Name: Tenn.
"""

import re
from datetime import datetime

from lxml.html import HtmlElement

from juriscraper.opinions.united_states.state import tenn
from juriscraper.OpinionSite import OpinionSite


class Site(tenn.Site):
    first_opinion_date = datetime(1996, 3, 14)
    days_interval = 30

    extract_from_text = OpinionSite.extract_from_text

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.tncourts.gov/courts/work-comp/opinions"
        self.section_selector = (
            "views-field views-field-field-opinions-case-brief"
        )

        # docket will always contain WC in the docket number
        self.docket_xpath = "text()[contains(., 'WC')]"

    def get_summary(self, section: HtmlElement) -> str:
        """Get the summary considering some XPath variations

        :param section: the html element containing the case summary
        :return the parsed summary, or an empty string
        """
        try:
            return super().get_summary(section)
        except IndexError:
            pass

        # last non empty floating text in the section
        summary = section.xpath("text()[normalize-space()]")[-1]
        # ensure we didn't pick up a judge field
        if " Judge:" not in summary:
            return summary

        # sometimes each line of the summary is inside a div
        # See 2022 case 'Sonney Summers v. RTR Transportation Services et al.'
        if containers := section.xpath("div/text()"):
            return re.sub(r"\s+", " ", " ".join(containers))

        return ""



"""Scraper for the D.C. Court of Appeals
CourtID: dc
Court Short Name: D.C.
Author: V. David Zvenyach
Date created:2014-02-21
History:
    - 2025-08-29: Update to OpinionSiteLinear and added extract_from_text method to get lower court info, Luis Manzur
    - 2026-03-02: Update url, Luis Manzur
"""

import re

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://www.dccourts.gov/court-of-appeals/opinions-and-memorandum-of-judgments"
        qualifier_no_opinions = (
            'not(contains(td[2]/span/text(), "NO OPINIONS"))'
        )
        qualifier_has_pdf_link = 'contains(.//td[1]/a/@href, ".pdf")'
        self.base_path = f"//table//tr[{qualifier_no_opinions} and {qualifier_has_pdf_link}]"
        self.should_have_results = True
        self.status = "Published"

    def _process_html(self):
        """Process the html and extract out the opinions

        :return: None
        """

        for row in self.html.xpath(self.base_path):
            docket = row.xpath("./td[1]/a/text()")[0].strip()
            url = row.xpath("./td[1]/a/@href")[0].strip()
            name = row.xpath("./td[2]/text()")[0].strip()
            date = row.xpath("./td[3]/text()")[0].strip()
            self.cases.append(
                {
                    "name": name,
                    "date": date,
                    "docket": docket,
                    "url": url,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """

        court_and_docket_pattern = re.compile(
            r"""
            (?:
               Appeals?\s+from\s+the\s+
               | On\s+Report\s+and\s+Recommendation\s+of\s+the\s+
               | On\s+Petition\s+for\s+Review\s+of\s+a\s+Decision(?:\s*\n*\s*|\s+)of\s+the\s+
               | On\s+Certified\s+Question\s+from\s+the\s+
               | On\s+Petition\s+for\s+Review\s+of\s+an\s+Order\s+of\s+the\s+
            )
            (.*?)
            \s*\n*\s*\(([^)]+)\)
            """,
            re.X | re.DOTALL | re.MULTILINE | re.IGNORECASE,
        )

        match = court_and_docket_pattern.search(scraped_text)
        if match:
            lower_court = re.sub(r"\s+", " ", match.group(1)).strip()
            lower_court_number = match.group(2).strip()

            return {
                "Docket": {
                    "appeal_from_str": lower_court,
                },
                "OriginatingCourtInformation": {
                    "docket_number": lower_court_number,
                },
            }

        return {}



"""
Scraper for the United States Bankruptcy Appellate Panel for the Ninth Circuit
CourtID: bap9
Court Short Name: 9th Cir. BAP
"""

import json
import re
from datetime import date, datetime, timedelta
from urllib.parse import urljoin

from juriscraper.AbstractSite import logger
from juriscraper.lib.auth_utils import generate_aws_sigv4_headers
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    query_url = "https://dynamodb.us-west-2.amazonaws.com/"
    days_interval = 31
    first_opinion_date = datetime(2005, 1, 6)
    lower_court_regex = re.compile(
        r"Appeals? from the (?P<lower_court>.+\n.+)"
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.base_url = "https://www.ca9.uscourts.gov/bap/"

        self.table = "bap"
        self.headers = {
            "Content-Type": "application/x-amz-json-1.1",
            "X-Amz-Target": "AWSCognitoIdentityService.GetCredentialsForIdentity",
        }
        self.params = {
            "IdentityId": "us-west-2:8d780f3b-d79c-c6c8-1125-e7a905da6b9b"
        }
        self.end_date = datetime.now()
        self.start_date = self.end_date - timedelta(days=self.days_interval)
        self.build_payload()
        self.url = "https://cognito-identity.us-west-2.amazonaws.com/"
        self.make_backscrape_iterable(kwargs)

    def build_payload(self):
        """Build query

        :return: query dict
        """
        expression_values = {
            ":start_date": {"S": self.start_date.strftime("%m")},
            ":year": {"S": self.start_date.strftime("%Y")},
        }
        filter_expression = (
            "#date_filed > :start_date and contains(#date_filed, :year)"
        )
        if date.month != 12:
            expression_values[":end_date"] = {
                "S": self.end_date.strftime("%m")
            }
            filter_expression = "#date_filed > :start_date and #date_filed < :end_date and contains(#date_filed, :year)"

        self.payload = json.dumps(
            {
                "TableName": self.table,
                "FilterExpression": filter_expression,
                "ExpressionAttributeNames": {"#date_filed": "date_filed"},
                "ExpressionAttributeValues": expression_values,
                "ReturnConsumedCapacity": "TOTAL",
            }
        )

    def _download(self):
        """Download data from DynamoDB for oral arguments.

        :return: The JSON response from DynamoDB containing oral argument records.
        """
        if self.test_mode_enabled():
            return json.load(open(self.url))

        sess = self.request["session"]

        # fetch for credentials
        res = sess.post(self.url, headers=self.headers, json=self.params)
        creds = res.json().get("Credentials")

        # fetch signed headers
        sig = generate_aws_sigv4_headers(self.payload, self.table, creds)

        logger.info(
            "Now downloading case page at: %s (params: %s)"
            % (self.url, self.payload)
        )

        # fetch bap table
        self.request["response"] = sess.post(
            self.query_url, headers=sig, data=self.payload
        )

        if self.save_response:
            self.save_response(self)

        self._post_process_response()
        return self._return_response_text_object()["Items"]

    def _process_html(self):
        """Process the HTML response and extract case details.

        Iterates over the items in the HTML response, extracts relevant
        case information, and appends it to the cases list.

        :return: None; updates self.cases with extracted case details.
        """
        for item in self.html:
            date_str = item.get("date_filed").get("S")
            try:
                d = datetime.strptime(date_str, "%m/%d/%Y")
                if d < self.start_date:
                    continue
                if d > self.end_date:
                    continue
            except ValueError:
                logger.debug("Skipping row with bad date data %s", item)
                continue

            slug = item.get("file_name").get("S")
            status = item.get("document_type").get("S")
            if status == "Unpublished Opinion":
                status = "Unpublished"
            else:
                status = "Published"
            self.cases.append(
                {
                    "name": f"In re: {item.get('debtor').get('S', '')}",
                    "date": item.get("date_filed").get("S"),
                    "url": urljoin("https://cdn.ca9.uscourts.gov", slug),
                    "docket": item.get("bap_num", {}).get("S"),
                    "status": status,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """
        if match := self.lower_court_regex.search(scraped_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()
            return {
                "Docket": {
                    "appeal_from_str": lower_court,
                }
            }

        return {}

    def _download_backwards(self, date: date) -> None:
        """Download cases for a specific date range.

        :param dates: A tuple containing start and end dates.
        :return: None; updates self.cases with cases from the specified date range.
        """
        start_str, end_str, _ = date

        # Parse start date or fall back to first_opinion_date
        if start_str:
            self.start_date = datetime.strptime(start_str, "%Y/%m/%d")
        else:
            self.start_date = self.first_opinion_date

        # Parse end date or fall back to now
        if end_str:
            self.end_date = datetime.strptime(end_str, "%Y/%m/%d")
        else:
            self.end_date = datetime.now()

        self.payload = json.dumps(
            {"TableName": self.table, "ReturnConsumedCapacity": "TOTAL"}
        )
        self.html = self._download()
        self._process_html()

    def make_backscrape_iterable(self, kwargs) -> None:
        """Make back scrape iterable

        :param kwargs: the back scraping params
        :return: None
        """
        self.start_date = kwargs.get(
            "backscrape_start", self.first_opinion_date
        )
        self.end_date = kwargs.get("backscrape_end", datetime.now())
        self.back_scrape_iterable = [(self.start_date, self.end_date, None)]



from juriscraper.opinions.united_states.state import calctapp_1st


class Site(calctapp_1st.Site):
    court_code = "S"
    division = ""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.courts.ca.gov/cms/npopinions.htm"
        self.court_id = self.__module__
        self.status = "Unpublished"
        self.should_have_results = True



"""Scraper for Dept of Justice Office of Legal Counsel
CourtID: bia
Court Short Name: Dept of Justice OLC
Author: William Palin
Reviewer:
Type:
History:
    2022-01-14: Created by William E. Palin
"""

from datetime import date, datetime
from urllib.parse import urlencode

from juriscraper.AbstractSite import logger
from juriscraper.lib.auth_utils import get_justice_dot_gov_auth_cookies
from juriscraper.lib.exceptions import UnexpectedContentTypeError
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    base_url = "https://www.justice.gov/olc/opinions"
    days_interval = 180
    first_opinion_date = datetime(1934, 3, 16)

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = self.base_url
        self.status = "Published"
        self.url = f"{self.base_url}?items_per_page=40"
        self.make_backscrape_iterable(kwargs)
        self.should_have_results = True

    def _process_html(self):
        for row in self.html.xpath(".//article"):
            name = row.xpath(".//h2")[0].text_content().strip()
            if not name:
                continue
            url = row.xpath(".//a/@href")[0]
            date_filed = row.xpath(".//time")[0].text_content()
            summary = row.xpath(".//p")[0].text_content()
            self.cases.append(
                {
                    "date": date_filed,
                    "name": name,
                    "url": url,
                    "summary": summary,
                    "docket": "",  # Docket numbers don't appear to exist.
                }
            )

    def _download_backwards(self, dates: tuple[date, date]) -> None:
        """Make custom date range request

        :param dates: (start_date, end_date) tuple
        :return None
        """
        logger.info("Backscraping for range %s %s", *dates)
        params = {
            "search_api_fulltext": "",
            "start_date": dates[0].strftime("%m/%d/%Y"),
            "end_date": dates[1].strftime("%m/%d/%Y"),
            "sort_by": "field_date",
            "items_per_page": "40",
        }
        self.url = f"{self.base_url}?{urlencode(params)}"
        self.html = self._download()
        self._process_html()

    def download_content(
        self, download_url, doctor_is_available=True, media_root=""
    ):
        """Overrides regular download_content to handle the
        "I am not a robot challenge". See #1724
        """
        try:
            return super().download_content(
                download_url, doctor_is_available, media_root
            )
        except UnexpectedContentTypeError as exc:
            # access HTML with JS variables to populate cookies
            html_text = exc.data["response"].text
            self.cookies = get_justice_dot_gov_auth_cookies(html_text)
            return super().download_content(
                download_url, doctor_is_available, media_root
            )



"""
Contact: Sara Velasquez, svelasquez@idcourts.net, 208-947-7501
History:
 - 2014-08-05, mlr: Updated.
 - 2015-06-19, mlr: Updated to simply the XPath expressions and to fix an OB1
   problem that was causing an InsanityError. The cause was nasty HTML in their
   page.
 - 2015-10-20, mlr: Updated due to new page in use.
 - 2015-10-23, mlr: Updated to handle annoying situation.
 - 2016-02-25 arderyp: Updated to catch "ORDER" (in addition to "Order") in download url text
 - 2024-12-30, grossir: updated to OpinionSiteLinear
 - 2025-07-09, luism: Updated to prevent wrongly formatted dates
 - 2025-08-29, luism: Added extract_from_text to get lower court information
"""

import re

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    # Skip rows that don't have  link in 4th cell with
    # either 'Opinion', 'Order', 'ORDER', or 'Amend' in
    # the link text
    path_conditional_anchor = (
        "a["
        'contains(.//text(), "Opinion") or '
        'contains(.//text(), "Order") or '
        'contains(.//text(), "ORDER") or '
        'contains(.//text(), "Amended")'
        "]"
    )

    # https://www.isc.idaho.gov/appeals-court/isc_civil
    base_url = "https://www.isc.idaho.gov/appeals-court/"
    url_part = "isc_civil"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = f"{self.base_url}{self.url_part}"
        self.court_id = self.__module__
        self.status = "Published"
        self.should_have_results = True

    def _process_html(self):
        row_xpath = f"//table//tr[.//{self.path_conditional_anchor}]"
        for row in self.html.xpath(row_xpath):
            url = self.get_opinion_url(row).replace("http://", "https://")

            # Sanitize case date, fix typo in date format
            date = row.xpath("string(td[1])").strip()
            if date == "Aug 22,2018":
                date = "Aug 22, 2018"

            self.cases.append(
                {
                    "date": date,
                    "docket": row.xpath("string(td[2])").strip(),
                    "name": row.xpath("string(td[3])").strip(),
                    "url": url,
                }
            )

    def get_opinion_url(self, row) -> str:
        """Get's the URL tagged as an Opinion, if possible

        We'll accept an order document if the opinion document
        is missing. Since each row can list multiple valid links,
        we will parse all acceptable links, take the opinion link
        if present, otherwise take the first acceptable link.

        :param row: the lxml object of the row
        :return: the document URL
        """

        for link in row.xpath("td[4]//a"):
            if "Opinion" in link.text_content().strip():
                return link.xpath("@href")[0]

        return row.xpath("td[4]//a/@href")[0]

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """
        pattern = re.compile(
            r"""
            Appeals?\s+from\s+the\s+
            (?P<lower_court>[^.]+?)
            (?=\s*\.)
            """,
            re.X | re.DOTALL,
        )

        judge_pattern = re.compile(
            r"County[.,]\s*(?P<judge>.+?), (?:Senior )?District Judge"
        )

        result = {}
        if match := pattern.search(scraped_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()
            if "County," in lower_court:
                lower_court = lower_court.split("County")[0] + "County"

            result["Docket"] = {
                "appeal_from_str": lower_court,
            }

            if match := judge_pattern.search(scraped_text):
                lower_court_judge = re.sub(
                    r"\s+", " ", match.group("judge")
                ).strip()
                result["OriginatingCourtInformation"] = {
                    "assigned_to_str": lower_court_judge
                }

        return result



import email
import pprint
import re
import sys
from datetime import datetime
from email.message import EmailMessage
from enum import Enum
from pathlib import Path
from typing import Optional, TypedDict, Union
from urllib.parse import parse_qs, urlparse

import requests
from lxml import html
from lxml.html import HtmlElement

from juriscraper.AbstractSite import logger
from juriscraper.lib.email_utils import parse_email_html
from juriscraper.lib.html_utils import clean_html
from juriscraper.lib.string_utils import clean_string, harmonize
from juriscraper.scotus import SCOTUSDocketReportHTML


class SCOTUSEmailType(Enum):
    """
    Enum for the different types of emails that can be received from SCOTUS.
    """

    INVALID = "invalid"
    CONFIRMATION = "confirmation"
    DOCKET_ENTRY = "docket_entry"


class SCOTUSNotificationEmail(TypedDict):
    """
    Schema for data extracted from SCOTUS notification emails.

    :ivar docket_number: The docket number of the case from the email.
    :ivar case_name: The name of the case from the email.
    :ivar entry_description: The description of the docket entry from the
    email.
    """

    docket_number: str
    case_name: str
    entry_description: str


class SCOTUSEmailData(TypedDict):
    """
    Schema for data extracted from SCOTUS emails.

    :ivar email_type: The type of email this data is extracted from.
    :ivar followup_url: The URL to perform the next action in the email flow
    (confirmation link for confirmation emails, case URL for docket entry
    emails). Must be set if the email type is valid and must be an empty string
    otherwise.
    :ivar email_datetime: The time (extracted from the "Date" email header) the
    email was sent.
    :ivar data: The data extracted from the email. Type corresponds to
    `email_type` (DOCKET_ENTRY: SCOTUSNotificationEmail; CONFIRMATION,
    UNKNOWN: None).
    """

    email_type: str
    followup_url: str
    email_datetime: datetime
    data: Optional[SCOTUSNotificationEmail]


class SCOTUSEmailHandlingResult(TypedDict):
    """
    Schema for result of handling SCOTUS emails.
    """

    email_type: str
    data: Union[dict[str, str], str]


class SCOTUSConfirmationResult(Enum):
    """
    Enum for the possible results of attempting to confirm a subscription to
    SCOTUS emails.
    """

    BadRequest = "bad_request"
    """Possibly: the request was malformed or missing required parameters.
    Unclear from SCOTUS website."""
    Failed = "failed"
    """Possibly: The confirmation request failed for some reason. Unclear from
    SCOTUS website."""
    NoVerify = "no_verify"
    """Unclear from SCOTUS website."""
    LinkExpired = "link_expired"
    """The confirmation link has expired."""
    NotFound = "not_found"
    """The email address associated with the confirmation link was not found
    (don't ask me how that could happen)."""
    AlreadySubscribed = "duplicate"
    """The email address is already subscribed to SCOTUS emails."""
    Success = "success"
    """The subscription confirmation was successful."""
    Unknown = "unknown"
    """Unknown error. Used when we can't determine the confirmation result.
    Should never appear in practice and exists only for future-proofing."""


class _SCOTUSConfirmationPageScraper:
    """
    Special-purpose scraper for parsing the subscription confirmation page for
    SCOTUS email updates.

    Should only be used by `SCOTUSEmail`.

    Mainly exists to make testing easier.

    :ivar court_id: Court ID for the scraper.
    :ivar tree: Parsed HTML tree.
    """

    def __init__(self, court_id: str = "scotus"):
        self.court_id = court_id
        self.tree: Optional[HtmlElement] = None

    @property
    def data(self) -> str:
        """
        Determine the result of our attempt to confirm our subscription.

        :return: Result of the confirmation attempt.
        """
        body_content = self.tree.find(".//div[@class='body-content']")
        script_tag = body_content.find(".//script")
        script = script_tag.text_content()
        # The confirmation page by default displays all response messages
        # and uses a (presumably) server-generated if/else chain with
        # conditions set to `true` or `false` to determine which message to
        # display. A better solution would be to either render the page with
        # JS enabled or to parse the script tag into an AST, but this works
        # for now.
        match = re.search(r"true\)\s\{([^}]+)", script)
        if match is None:
            return SCOTUSConfirmationResult.NoVerify.value
        statement_body = match.group(1)
        visibility_calls = [
            line.strip() for line in statement_body.split("\n")[1:-1]
        ]
        visibility_re = re.compile(r"^\$\(\'#(.+)\'\)\.(show|hide)\(\);$")
        visibility_matches = [
            visibility_re.match(call) for call in visibility_calls
        ]
        confirmation_status = next(
            visibility_match.group(1)[3:].lower()
            for visibility_match in visibility_matches
            if visibility_match is not None
            and visibility_match.group(2) == "show"
        )

        if confirmation_status == "badrequest":
            return SCOTUSConfirmationResult.BadRequest.value
        if confirmation_status == "fail":
            return SCOTUSConfirmationResult.Failed.value
        if confirmation_status == "noverifymessage":
            return SCOTUSConfirmationResult.NoVerify.value
        if confirmation_status == "linkexpired":
            return SCOTUSConfirmationResult.LinkExpired.value
        if confirmation_status == "notfound":
            return SCOTUSConfirmationResult.NotFound.value
        if confirmation_status == "duplicate":
            return SCOTUSConfirmationResult.AlreadySubscribed.value
        if confirmation_status == "ok":
            return SCOTUSConfirmationResult.Success.value
        return SCOTUSConfirmationResult.Unknown.value

    def _parse_text(self, text: str) -> None:
        """Parse raw HTML and store a lxml tree."""
        self.tree = html.fromstring(clean_html(text))


class SCOTUSEmail:
    """Parse SCOTUS docket notification email."""

    TITLE_REGEX = re.compile(r"^A new docket entry, \"(.+?)\" has been added")
    DOCKET_ENTRY_SUBJECT_REGEX = re.compile(
        r"^Supreme Court Electronic Filing System$"
    )
    CONFIRMATION_SUBJECT_REGEX = re.compile(
        r"^Supreme Court Case Notification .+$"
    )

    def __init__(self, court_id: str = "scotus"):
        self.court_id: str = court_id
        self.tree: Optional[HtmlElement] = None
        self.message: Optional[EmailMessage] = None
        self.email_type: SCOTUSEmailType = SCOTUSEmailType.INVALID

    @property
    def data(self) -> SCOTUSEmailData:
        """Extract all the data in the email.

        If the email is a docket update email, the output will be formatted:

        :return: Data extracted from the email. The `data` key will be
        `None` if the email type is unknown or parsing fails.
        """

        if self.email_type == SCOTUSEmailType.DOCKET_ENTRY:
            followup_url = self._parse_first_link()
            data = SCOTUSNotificationEmail(
                docket_number=self._parse_docket_number(),
                case_name=self._parse_case_name(),
                entry_description=self._parse_filing_name(),
            )
        elif self.email_type == SCOTUSEmailType.CONFIRMATION:
            followup_url = self._parse_first_link()
            data = None
        else:
            followup_url = ""
            data = None

        return SCOTUSEmailData(
            email_type=self.email_type.value,
            followup_url=followup_url,
            email_datetime=self._parse_datetime(),
            data=data,
        )

    def handle_email(self, timeout: float = 10.0) -> SCOTUSEmailHandlingResult:
        """
        Handle next steps in email processing.

        If the email is a docket update, request the docket page and return the
        parsed data.

        If the email is a confirmation email, attempt to confirm the
        subscription and return the parsed result.

        :param timeout: Timeout for the HTTP request in seconds.

        :return: Result of handling the email.
        """
        if self.email_type == SCOTUSEmailType.INVALID:
            raise TypeError("Unknown email type.")

        # The `followup_url` property is guaranteed to be present if the
        # email type is valid.
        response = requests.get(
            self.data["followup_url"],
            headers={"User-Agent": "Free Law Project"},
            timeout=timeout,
        )
        response.raise_for_status()

        if self.email_type == SCOTUSEmailType.DOCKET_ENTRY:
            scotus_docket_report_html = SCOTUSDocketReportHTML(self.court_id)
            scotus_docket_report_html._parse_text(response.text)
            return SCOTUSEmailHandlingResult(
                email_type=self.email_type.value,
                data=scotus_docket_report_html.data,
            )

        # The only remaining possibility is a confirmation email.
        scotus_confirmation_page_scraper = _SCOTUSConfirmationPageScraper(
            self.court_id
        )
        scotus_confirmation_page_scraper._parse_text(response.text)
        return SCOTUSEmailHandlingResult(
            email_type=self.email_type.value,
            data=scotus_confirmation_page_scraper.data,
        )

    def _determine_email_type(self) -> SCOTUSEmailType:
        """Determine the type of the email (docket update/confirmation) based
        on the subject line. If the subject line does not match any known
        patterns, return `EmailType.INVALID`."""
        subject = self.message.get("Subject", failobj="")

        if self.DOCKET_ENTRY_SUBJECT_REGEX.match(subject) is not None:
            return SCOTUSEmailType.DOCKET_ENTRY
        elif self.CONFIRMATION_SUBJECT_REGEX.match(subject) is not None:
            return SCOTUSEmailType.CONFIRMATION
        logger.error(
            "Email subject did not match known patterns. Subject: %s", subject
        )
        return SCOTUSEmailType.INVALID

    def _parse_text(self, text: str) -> None:
        """Extract and store the first part of the email with the
        "Content-Type" header set to "text/html" and store a parsed HTML
        tree. Assumes that there will always be an email part with this
        content type. If lxml parsing fails, return `None` and log an
        appropriate error.

        :param text: The raw email text.

        :return: None
        """
        self.message = email.message_from_string(text)
        email_type = self._determine_email_type()

        if email_type == SCOTUSEmailType.INVALID:
            return

        self.tree = parse_email_html(text)
        if self.tree is None:
            return

        n_anchors = sum(1 for _ in self.tree.iterfind(".//a"))

        if email_type == SCOTUSEmailType.DOCKET_ENTRY:
            if self.message.get("Date") is None:
                logger.error("Unable to find 'Date' header in email")
                return

            if n_anchors < 2:
                logger.error(
                    "Unable to find at least two links in email body (should "
                    "be case link and unsubscribe link); found %d",
                    n_anchors,
                )
                return

            text = self.tree.text_content()
            if self.TITLE_REGEX.match(text) is None:
                logger.error(
                    "Unable to find match for docket entry title regex in "
                    "email"
                )
                return

            self.email_type = email_type
        elif email_type == SCOTUSEmailType.CONFIRMATION:
            if n_anchors != 1:
                logger.error(
                    "Incorrect number of links in email body. Should be "
                    "exactly one (confirmation link); found %d",
                    n_anchors,
                )
                return
            self.email_type = email_type

    def _parse_datetime(self) -> Optional[datetime]:
        """Extract the "Date" header in the notification email message into a
        `datetime`, returning `None` if the header is absent or parsing fails
        and logging an appropriate error.

        :return: `datetime` or `None` if unable to parse the "Date" header.
        """
        message_date = self.message.get("Date")

        try:
            return datetime.strptime(message_date, "%a, %d %b %Y %H:%M:%S %z")
        except ValueError:
            logger.error(
                "Unable to parse 'Date' header of email (value is '%s')",
                message_date,
            )
            return None

    def _parse_filing_name(self) -> str:
        """Extract the docket entry title from the email body using a regex
        and clean it with the `clean_string` method. Return `None` and log
        an error if the regex fails to match the email.

        :return: Clean docket entry title.
        """
        text = self.tree.text_content()
        match = self.TITLE_REGEX.match(text)

        return clean_string(match.group(1))

    def _parse_case_name(self) -> str:
        """Extract the case name from the first `<a>` tag in the email body
        and uses the `harmonize` method to clean it.

        :return: Cleaned case name.
        """
        return harmonize(self.tree.findtext(".//a"))

    def _parse_docket_number(self) -> str:
        """Extract the docket number using the `href` attribute from the first
        `<a>` tag in the email body.

        :return: Docket number.
        """
        file = parse_qs(urlparse(self.tree.find(".//a").get("href")).query)[
            "filename"
        ][0]
        docket_number = Path(file).stem
        return clean_string(docket_number)

    def _parse_first_link(self) -> str:
        """Extract the `href` attribute from the first `<a>` tag in the email
        body.
        """
        return self.tree.find(".//a").get("href")


def _main():
    """Parse a local email file and pretty print extracted data.

    :return: None
    """
    if len(sys.argv) != 2:
        print("Usage: python -m juriscraper.scotus_email <filepath>")
        print("Please provide a path to an email file to parse.")
        sys.exit(1)

    report = SCOTUSEmail()
    filepath = sys.argv[1]
    print(f"Parsing email file at {filepath}")
    with open(filepath, encoding="utf-8") as f:
        text = f.read()

    report._parse_text(text)
    pprint.pprint(report.data, indent=2)


if __name__ == "__main__":
    _main()



# Scraper for Texas Business Court
# CourtID: texbizct
# Author: Luis Manzur
# History:
#  - 2025-07-29: Created by Luis Manzur
import re
from typing import Optional

from dateutil import parser

from juriscraper.lib.log_tools import make_default_logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear

logger = make_default_logger()


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = "https://www.txcourts.gov/businesscourt/opinions/"
        self.court_id = self.__module__
        self.should_have_results = True
        self.status = "Published"

    def _process_html(self) -> None:
        """Parses the HTML content

        :return None
        """

        links = self.html.xpath('//div[@class="panel-content"]//a')
        for link in links:
            title = link.get("title")
            short_title = link.text_content()

            # Attempt to split the short_title by the last comma to extract name and citation.
            # If that fails, try splitting the title instead.
            # If both attempts fail, assign the entire title to name and leave citation empty.
            try:
                name, citation = short_title.rsplit(", ", 1)
            except (ValueError, AttributeError):
                try:
                    citation = title.rsplit(", ", 1)[-1]
                    name = short_title
                except (ValueError, AttributeError):
                    name, citation = short_title, ""

            url = link.get("href")
            date = self._get_approximate_date(url)
            raw_docket = title.split()[0]
            docket = raw_docket.strip(",").replace("--", "-")

            summary = link.xpath("../following-sibling::ul/li/text()")
            summary = " ".join(summary)

            self.cases.append(
                {
                    "docket": docket,
                    "name": name.split("(", 1)[0].strip(),
                    "citation": citation.split("(", 1)[0].strip(),
                    "url": url,
                    "date": date,
                    "date_filed_is_approximate": True,
                    "summary": summary,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extracts date filed from the scraped text.

        :param scraped_text: The text content of the opinion.
        :return: A dictionary with case details.
        """

        # Get date from stamp on first page
        pattern = r"\b(\d{1,2}/\d{1,2}/\d{4})\b"
        dates = re.findall(pattern, scraped_text[:500])
        if dates:
            date_filed = dates[0]
        else:
            # if no stamp grab from signature block
            date_filed_pattern = (
                r"(?:\:|Signed)\s*([A-Za-z]+\s\d{1,2},\s\d{4})"
            )
            date_filed_match = re.search(date_filed_pattern, scraped_text)
            date_filed = (
                date_filed_match.group(1) if date_filed_match else None
            )

        if date_filed:
            dt = parser.parse(date_filed)
            date_string = dt.strftime("%Y-%m-%d")
            return {
                "OpinionCluster": {
                    "date_filed": date_string,
                    "date_filed_is_approximate": False,
                }
            }
        return {}

    def _get_approximate_date(self, url: str) -> Optional[str]:
        """Get Approximate date from head request

        :param url: The pdf url
        :return: The date the file was last touched
        """
        if self.test_mode_enabled():
            return "2025-01-01"
        resp = self.request["session"].head(
            url, allow_redirects=True, timeout=30
        )
        lm = resp.headers.get("Last-Modified")
        dt = parser.parse(lm)
        return dt.strftime("%Y-%m-%d")



from juriscraper.opinions.united_states.state import wash


class Site(wash.Site):
    crt_level = "C"
    pub_status = "PUB"
    name_td_index = 4
    type_td_index = 5

    # Example URL
    # https://www.courts.wa.gov/opinions/index.cfm?fa=opinions.byYear&fileYear=2025&crtLevel=C&pubStatus=PUB

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.status = "Published"



from abc import ABC, abstractmethod
from typing import Generic, TypeVar

T = TypeVar("T")


class AbstractParser(Generic[T], ABC):
    """Abstract base class, which all scrapers should inherit from."""

    @abstractmethod
    def __init__(self, court_id: str):
        self.court_id: str = court_id

    @abstractmethod
    def _parse_text(self, text: str) -> None:
        """Ingest some arbitrary text and prepare it to be parsed by the data() method."""
        pass

    @property
    @abstractmethod
    def data(self) -> T:
        """Parse the last text ingested by _parse_text()."""
        pass



"""
Scraper for Pennsylvania Supreme Court
CourtID: pa
Court Short Name: pa
"""

import re
from datetime import date, datetime, timedelta
from urllib.parse import urlencode

from juriscraper.AbstractSite import logger
from juriscraper.ClusterSite import ClusterSite
from juriscraper.lib.type_utils import OpinionType


class Site(ClusterSite):
    court = "Supreme"
    base_url = "https://www.pacourts.us/api/opinion?"
    document_url = "https://www.pacourts.us/assets/opinions/{}/out/{}"
    days_interval = 1
    api_dt_format = "%Y-%m-%dT00:00:00-05:00"
    first_opinion_date = datetime(1998, 4, 27)
    judge_key = "AuthorCode"
    regional_cite_regex = re.compile(r"\d{1,3} A\.3d \d+")
    post_type_key = "PostingTypeId"

    def __init__(self, *args, **kwargs):
        """About postTypes values in self.params


        mo = majority opinion
        pco = per curiam order
        co = concurring opinion
        cs = concurring statement
        cds = concurring and dissenting statement
        do = dissenting opinion
        ds = dissenting statement
        oaj = opinion announcing judgement of the court
        rv = Evenly Divided Court, Reversal
        dedc = Dismissal, Evenly Divided Court
        """
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.regex = re.compile(r"(.*)(?:[,-]?\s+Nos?\.)(.*)")
        self.status = "Published"

        now = datetime.now()
        start = now - timedelta(days=1)
        self.params = {
            "startDate": start.strftime(self.api_dt_format),
            "endDate": now.strftime(self.api_dt_format),
            "courtType": self.court,
            "postTypes": "cd,cds,co,cs,dedc,do,ds,mo,oaj,pco,rv,sd",
            "sortDirection": "-1",
        }
        self.url = f"{self.base_url}{urlencode(self.params)}"
        self.make_backscrape_iterable(kwargs)

    def _process_html(self) -> None:
        """Parses data into case dictionaries

        Note that pages returned have no pagination

        :return: None
        """
        json_response = self.html

        for cluster in json_response["Items"]:
            title = cluster["Caption"]
            name, docket = self.parse_case_title(title)
            # A.3d cites seem to exist only for pasuperct
            cite = ""
            if cite_match := self.regional_cite_regex.search(title):
                cite = cite_match.group(0)

            parsed_cluster = {
                "date": cluster["DispositionDate"].split("T")[0],
                "name": name,
                "docket": docket,
                "citation": cite,
                "sub_opinions": [],
                "judge": "",
            }

            for op in cluster["Postings"]:
                post_type = op.get("PostType", {}).get(self.post_type_key, "")
                opinion_type = self.get_type(post_type, cluster)

                status = self.get_status(op)
                if not parsed_cluster.get("status"):
                    parsed_cluster["status"] = status

                per_curiam = False
                author_str = ""
                if op["Author"]:
                    author_str = self.clean_judge(op["Author"][self.judge_key])
                    if author_str.lower() == "per curiam":
                        author_str = ""
                        per_curiam = True
                    else:
                        if author_str not in parsed_cluster["judge"]:
                            parsed_cluster["judge"] += f"; {author_str}"

                if opinion_type == OpinionType.UNANIMOUS:
                    per_curiam = True
                    author_str = ""

                url = self.document_url.format(self.court, op["FileName"])
                parsed_cluster["sub_opinions"].append(
                    {
                        "url": url,
                        "author": author_str,
                        "per_curiam": per_curiam,
                        "type": opinion_type.value,
                    }
                )

            self.cases.append(parsed_cluster)

        if not self.test_mode_enabled() and json_response.get("HasNext"):
            next_page = json_response["PageNumber"] + 1
            logger.info("Paginating to page %s", next_page)
            self.params["pageNumber"] = next_page
            self.url = f"{self.base_url}{urlencode(self.params)}"
            self.html = self._download()
            self._process_html()

    def parse_case_title(self, title: str) -> tuple[str, str]:
        """Separates case_name and docket_number from case string

        :param title: string from the source

        :return: A tuple with the case name and docket number
        """
        search = self.regex.search(title)
        if search:
            name = search.group(1)
            docket = search.group(2)
        else:
            name = title
            docket = ""
        return name, docket

    def get_status(self, op: dict) -> str:
        """Get status from opinion json.
        Inheriting classes have status data

        :param op: opinion json
        :return: parsed status
        """
        return self.status

    def clean_judge(self, author_str: str) -> str:
        """Cleans judge name. `pa` has a different format than
        `pasuperct` and `pacommwct`
        """
        return author_str

    def get_type(self, post_type: str, cluster: dict) -> str:
        """Parse the PostingType into one of our Opinion.type values

        :param post_type: type as returned by the source
        :param cluster: the cluster data returned by the source
        :return: the parsed type
        """
        is_single_opinion = len(cluster["Postings"]) == 1

        if (
            post_type in ["Opinion Per Curiam", "Per Curiam Order"]
            and is_single_opinion
        ):
            type = OpinionType.UNANIMOUS
        elif post_type in [
            "Concurring Opinion",
            "Concurrent Statement",
            "Concurring Statement",
            "Concurring Memorandum",
            "Concurring Memorandum Statement",
        ]:
            type = OpinionType.CONCURRENCE
        elif post_type in [
            "Dissenting Opinion",
            "Dissenting Statement",
            "Dissenting Memorandum",
            "Dissenting Memorandum Statement",
        ]:
            type = OpinionType.DISSENT
        elif post_type in [
            "Concurring and Dissenting Opinion",
            "Concurring and Dissenting Statement",
            "Concurring and Dissenting Memorandum",
            "Concurring and Dissenting Memorandum Statement",
        ]:
            type = OpinionType.CONCURRING_IN_PART_AND_DISSENTING_IN_PART
        elif post_type == "Complete Opinion":
            type = OpinionType.COMBINED
        elif (
            post_type
            in [
                "Majority Opinion",
                "Opinion Announcing the Judgment of the Court",
                "Opinion",
                "Memorandum",
            ]
            or not is_single_opinion
        ):
            type = OpinionType.MAJORITY
        else:
            # default value in CL
            type = OpinionType.COMBINED

        return type

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract citations from text

        Be careful with citations referring to other opinions that are
        mentioned before the actual citation

        See, for example:
        https://ojd.contentdm.oclc.org/digital/api/collection/p17027coll5/id/28946/download
        """
        # For each line, keep only the part after the last colon
        cleaned_text = "\n".join(
            line.rsplit(":", 1)[-1] if ":" in line else line
            for line in scraped_text.splitlines()
        )

        pattern = re.compile(
            r"""
                    (?:
                        review\s+from\s+the\s+
                        |Order\s+of(?:\s+the)?\s+
                        |Appeal\s+from\s+the(?:\s[Oo]rder\s+of\s+the)?\s+
                     )
                    (?P<lower_court>.*?)(?=\s*(?:\.|at|entered|Order|,))
                    """,
            re.X | re.DOTALL,
        )
        result = {}
        if match := pattern.search(cleaned_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()

            if lower_court in ["Superior Court", "Commonwealth Court"]:
                lower_court = "Pennsylvania " + lower_court

            result["Docket"] = {"appeal_from_str": lower_court}

        return result

    def _download_backwards(self, dates: tuple[date, date]) -> None:
        """Modify GET querystring for desired date range

        :param dates: (start_date, end_date) tuple
        :return None
        """
        start, end = dates
        self.params["startDate"] = start.strftime(self.api_dt_format)
        self.params["endDate"] = end.strftime(self.api_dt_format)
        self.url = f"{self.base_url}{urlencode(self.params)}"
        logger.info("Backscraping for range %s %s", *dates)
        self.html = self._download()
        self._process_html()



from typing import Optional

from dateutil import parser

from juriscraper.AbstractSite import logger
from juriscraper.lib.exceptions import InsanityException
from juriscraper.lib.utils import (
    clean_attribute,
    sanity_check_case_names,
    sanity_check_dates,
    sanity_check_opinion_types,
)
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class ClusterSite(OpinionSiteLinear):
    """
    Keeps an interface compatible with AbstractSite

    Main changes
    + uses a dict representation of a case; instead of the attribute list
        representation. We override
            __iter__
            __getitem__
            __len__
            important steps inside parse
                _clean_attributes
                _date_sort
                _check_sanity
                _get_case_name_shorts


    + each returned dict now represent a opinion cluster; no longer a plain opinion
        attribute
        Each opinion inside the cluster will have different values for
            - authors
            - joined by
            - types
            - download_urls

    For now, we keep the plural nomenclature 'authors', 'types', etc. This will
    make CL changes easier
    """

    short_to_full_attribute_name = {
        "name": "case_names",
        "date": "case_dates",
        "status": "precedential_statuses",
        "docket": "docket_numbers",
        "judge": "judges",
        "citation": "citations",
        "parallel_citation": "parallel_citations",
        "summary": "summaries",
        "child_courts": "child_courts",
        "disposition": "dispositions",
        "nature_of_suit": "nature_of_suit",
        "other_date": "other_dates",
        "attorney": "attorneys",
        "headnote": "headnotes",
        # lower court info
        "lower_court": "lower_courts",
        "lower_court_id": "lower_court_ids",
        "lower_court_judge": "lower_court_judges",
        "lower_court_number": "lower_court_numbers",
        "court_reporter": "court_reporter",
        # special field to hold the opinion list
        "sub_opinions": "sub_opinions",
        # opinion fields
        "author": "authors",
        "per_curiam": "per_curiam",
        "joined_by": "joined_by",
        "type": "types",
        "url": "download_urls",
        # present in OpinionSite but unused in regular case law flow
        "docket_document_number": "docket_document_numbers",
        "docket_attachment_number": "docket_attachment_numbers",
        "cause": "causes",
        "division": "divisions",
        "adversary_number": "adversary_numbers",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.all_attributes = self._all_attrs
        # doing this will help us re-use `AbstractSite.parse``
        self._all_attrs = []

        # if set,  allow clustering when dates are not exactly the same
        self.cluster_by_date_max_days = 0

    def __iter__(self):
        yield from self.cases

    def __str__(self):
        return self.cases

    def __getitem__(self, i):
        return self.cases[i]

    def __len__(self):
        return len(self.cases)

    def normalize_attribute_name(self, name: str) -> str:
        """If short-hand attribute names were used, normalize them into full form
        supported by AbstractSite

        :param name: an possible attribute name
        :return: the normalized name
        :raises: ValueError if the input name has no normalized form
        """
        if name in self.all_attributes:
            return name

        if self.short_to_full_attribute_name.get(name):
            return self.short_to_full_attribute_name[name]

        raise ValueError(f"Invalid normalized name '{name}'")

    def _clean_attributes(self):
        """Perform name normalization and standard attribute cleaning.
        Handles nested opinion objects
        """
        cases = []
        for case in self.cases:
            cleaned_case = {}
            for key, value in case.items():
                normalized_name = self.normalize_attribute_name(key)

                # handle sub opinion lists
                if isinstance(value, list):
                    sub_opinion_list = []
                    for sub_opinion in value:
                        sub_op = {}
                        for k, v in sub_opinion.items():
                            normalized_opinion_key = (
                                self.normalize_attribute_name(k)
                            )
                            sub_op[normalized_opinion_key] = clean_attribute(
                                normalized_opinion_key, v
                            )

                        sub_opinion_list.append(sub_op)
                    cleaned_case[normalized_name] = sub_opinion_list
                else:
                    cleaned_case[normalized_name] = clean_attribute(
                        normalized_name, value
                    )

            cases.append(cleaned_case)

        self.cases = cases

    def _date_sort(self):
        """Orders cases by values of _req_attributes, beggining with date and name"""
        self.cases.sort(
            key=lambda case: tuple(
                case.get(attr, "") for attr in self._req_attrs
            ),
            reverse=True,
        )

    def _post_parse(self):
        """Let's use this hook to add default values for required attributes"""
        for case in self.cases:
            if not case.get("case_name_short"):
                case["case_name_shorts"] = self.cnt.make_case_name_short(
                    case["case_names"]
                )
            if not case.get("date_filed_is_approximate"):
                case["date_filed_is_approximate"] = False
            if not case.get("blocked_statuses"):
                case["blocked_statuses"] = False
            if not case.get("precedential_statuses"):
                if self.status:
                    case["precedential_statuses"] = self.status
                else:
                    raise InsanityException(
                        f"No value for `precedential_statuses` for case {case}"
                    )

    def _make_hash(self):
        """Populate the expected self.cases list; just before calling the parent
        _make_hash
        """
        self.case_names = [i["case_names"] for i in self.cases]
        super()._make_hash()

    def _check_sanity(self):
        """Perform standard validity checks. Handles nested objects

        Checks that
        - results exist
        - each case dictionary has all of the required attributes
        - case names are valid
        - dates are valid
        """
        if len(self.cases) == 0:
            self.no_results_warning()
            return

        for case in self.cases:
            for field in self._req_attrs:
                if case.get(field) is None:
                    # should only apply to `download_urls` for nested sub opinions
                    # but let's write the general case
                    if case.get("sub_opinions") and all(
                        op.get(field) is not None
                        for op in case["sub_opinions"]
                    ):
                        continue
                    raise InsanityException(
                        f"Case {case} has no value for required '{field}'"
                    )

        case_names = []
        dates_and_names = []
        for i in self.cases:
            case_names.append(i["case_names"])
            dates_and_names.append(
                (
                    i["case_dates"],
                    i.get("date_filed_is_approximate"),
                    i.get("case_names"),
                )
            )

        sanity_check_case_names(case_names)
        sanity_check_dates(dates_and_names, self.court_id)

        # Check opinion type rules for clusters with sub_opinions
        for case in self.cases:
            if case.get("sub_opinions"):
                sanity_check_opinion_types(case["sub_opinions"])

        logger.info(
            f"{self.court_id}: Successfully found {len(self.cases)} items."
        )

    def validate_cluster_dates(
        self, datestring1: str, datestring2: str
    ) -> bool:
        """Check that 2 dates are close enough to cluster the opinions

        By default, it will check that the date strings are exactly the same

        If the instance attribute `cluster_by_date_max_days` is set to a value
        different from 0, it will try to parse the dates and check if the dates
        are close enough

        :param datestring1: a date string
        :param datestring2: another date strin
        :return: True if the datestrigs or their dates are close enough
        """
        if self.cluster_by_date_max_days:
            dt1 = parser.parse(datestring1)
            dt2 = parser.parse(datestring2)
            return abs((dt1 - dt2).days) < self.cluster_by_date_max_days
        else:
            return datestring1 == datestring2

    def cluster_opinions(
        self, case_dict: dict, possible_clusters: list[dict]
    ) -> Optional[dict]:
        """Try to cluster current opinion with previous opinions.
        To cluster, opinions should have the same
        - date
        - case name
        - docket number

        The caller should take care to propagate any metadata from the
        `case_dict` to the cluster proper

        :param case_dict: the case dict we will atempt to merge into any of the
            possible clusters
        :param possible_clusters: the existing clusters we will try to compare

        :return: None if we couldn't find a cluster
            the cluster dict if the current case dict was clustered
        """
        if not possible_clusters:
            return

        opinion_fields = ["type", "url", "per_curiam", "author", "joined_by"]

        for candidate_cluster in possible_clusters:
            # all of these should be the same for this to be considered a cluster
            if (
                case_dict["name"] != candidate_cluster["name"]
                or case_dict["docket"] != candidate_cluster["docket"]
                or not self.validate_cluster_dates(
                    case_dict["date"], candidate_cluster["date"]
                )
            ):
                continue

            # if the sub_opinions list does not exist yet, build it from the
            # existing candidate_cluster
            sub_opinions = candidate_cluster.get("sub_opinions")
            if not sub_opinions:
                candidate_cluster["sub_opinions"] = [
                    {
                        key: candidate_cluster.pop(key)
                        for key in opinion_fields
                        if key in candidate_cluster
                    }
                ]

            # add the new opinion to the cluster
            candidate_cluster["sub_opinions"].append(
                {
                    key: case_dict.pop(key)
                    for key in opinion_fields
                    if key in case_dict
                }
            )

            return candidate_cluster

        return



"""Scraper for Second Circuit
CourtID: ca2
Author: MLR
Reviewer: MLR
Contact:
  Webmaster@ca2.uscourts.gov ('Calendar Team' handles updates, and is responsive)
  ecfhelpdesk@ca2.uscourts.gov
  Shane_Clouden@ca2.uscourts.gov
History:
  2016-09-09: Created by MLR
  2023-11-21: Fixed by flooie
  2023-12-11: Fixed by quevon24
  2026-01-21:Fixed by Luis-manzur
"""

import re
from datetime import datetime
from urllib.parse import urljoin

from juriscraper.AbstractSite import logger
from juriscraper.OralArgumentSiteLinear import OralArgumentSiteLinear


class Site(OralArgumentSiteLinear):
    base_url = "http://ww3.ca2.uscourts.gov"
    docket_regex = re.compile(r"(?<!\d-)\d{2}-\d{2,4}(?!-\d)")
    date_regex = re.compile(r"^\d{1,2}-\d{1,2}-\d{2}$")

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = f"{self.base_url}/decisions"

        self.method = "POST"
        self.expected_content_types = ["audio/mpeg3"]
        self.parameters = self.request["parameters"]["params"] = {
            "IW_SORT": "-DATE",
            "IW_BATCHSIZE": "50",
            "IW_FILTER_DATE_BEFORE": "",
            "IW_FILTER_DATE_AFTER": "NaNNaNNaN",
            "IW_FIELD_TEXT": "*",
            "IW_DATABASE": "Oral Args",
            "opinion": "*",
        }

        self.back_scrape_iterable = ["placeholder"]

    def _process_html(self):
        for row in self.html.xpath("//table[@border='1']"):
            link = row.xpath(".//a")[0].get("href")

            if ".mp3" not in link:
                logger.info("Skipping row without mp3 link: %s", link)
                continue  # skip bad data
            url = urljoin(self.base_url, link)

            docket = row.xpath(".//a/nobr/text()")[0]

            try:  # Row may have empty date column
                name, date = row.xpath(".//td/text()")
            except ValueError:
                name = row.xpath(".//td/text()")[0]
                date = (
                    self.cases[-1]["date"]
                    if self.cases
                    else datetime.now().strftime("%-m-%d-%y")
                )  # use last case date

            if not self.docket_regex.search(
                docket
            ):  # Row may have mixed columns
                if self.date_regex.search(docket):
                    date = docket
                    docket = self.docket_regex.search(link).group(0)
                else:
                    docket = name
                    name = row.xpath(".//a/nobr/text()")[0]

            if docket == "GMT20230614-175136_Recording":
                continue  # skip bad data

            self.cases.append(
                {
                    "docket": docket,
                    "url": url,
                    "name": name,
                    "date": date,
                }
            )

    def _download_backwards(self, d):
        # This backscraper is hardcoded, but it will
        # work for the set number of pages
        self.method = "GET"
        for i in range(1, 33):
            self.url = f"{self.base_url}/decisions/isysquery/19342efa-bb49-4684-a054-875324b58eb9/{(i * 10) - 9}-{i * 10}/list/"
            self.html = self._download()
            self._process_html()



"""Scraper for Eleventh Circuit of Appeals
CourtID: ca11
Court Short Name: ca11
Author: Jon Andersen
Reviewer: mlr
Date created: 28 Aug 2018
"""

from juriscraper.OralArgumentSiteLinear import OralArgumentSiteLinear


class Site(OralArgumentSiteLinear):
    base_url = "https://www.ca11.uscourts.gov/oral-argument-recordings"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = self.base_url
        self.back_scrape_iterable = list(range(0, 52))

    def _process_html(self):
        for row in self.html.xpath("//tr[not(th)]"):
            # normalize docket numbers
            # get rid of "consolidated with" text
            # parse docket numbers like docketnum1 & docketnum2
            # also handle docketnum1\ndocketnum2\ndocketnum3
            # Return comma joined string like docketnum1, docketnum2
            docket = row.xpath("td[1]/text()")[0]
            docket = ", ".join(
                docket.strip()
                .upper()
                .replace("&", "\n")
                .replace(" AND ", "\n")
                .replace(",", "\n")
                .replace("CONS. WITH", "\n")
                .replace("CONSOLIDATED WITH", "\n")
                .split()
            )
            name = row.xpath("td[2]/text()")[0]
            date = row.xpath("td[3]/text()")[0]
            url = row.xpath("td[5]//@href")[0]
            self.cases.append(
                {"name": name, "url": url, "date": date, "docket": docket}
            )

    def _download_backwards(self, i):
        self.url = f"{self.base_url}?page={i}"
        self.html = self._download()
        self._process_html()



"""
Scraper for Indiana Tax Court
CourtID: indtc
Court Short Name: Ind. Tax.
Auth: Jon Andersen <janderse@gmail.com>
Reviewer: mlr
History:
    2014-09-03: Created by Jon Andersen
"""

import re

from juriscraper.opinions.united_states.state import ind


class Site(ind.Site):
    page_court_id = "9550"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.court_name = "Tax Court"
        self.status = "Unknown"

    def extract_from_text(self, scraped_text: str):
        """Extract publication status

        :param scraped_text: Text of scraped content
        :return: metadata
        """
        if re.findall(r"\bNOT FOR PUBLICATION\b", scraped_text):
            status = "Unpublished"
        elif re.findall(r"\bFOR PUBLICATION\b", scraped_text):
            status = "Published"
        else:
            status = "Unknown"
        metadata = {
            "OpinionCluster": {
                "precedential_status": status,
            },
        }
        return metadata



"""
Scraper for Kansas Supreme Court
CourtID: kan
Court Short Name: Kansas
Author: William Palin
Court Contact:
History:
 - 2025-06-09: Created.
Notes: This is a selenium scraper and must start or redirect from somewhere else
"""

import re

from lxml.html import fromstring

from juriscraper.OpinionSiteLinearWebDriven import OpinionSiteLinearWebDriven


class Site(OpinionSiteLinearWebDriven):
    link_xp = '//div[@class="col-md-6"][1]/ul/li'

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = "https://free.law"
        self.uses_selenium = True
        self.should_have_results = True

    def _process_html(self):
        """Process with selenium

        :return: None
        """
        if not self.test_mode_enabled():
            self.initiate_webdriven_session()
            self.url = "https://searchdro.kscourts.gov/Data/decisionsList"
            self.webdriver.get(self.url)
            self.html = fromstring(self.webdriver.page_source)
        else:
            with open(self.url) as f:
                self.html = fromstring(f.read())

        links = self.html.xpath(self.link_xp)
        for link in links:
            # Get previous date
            date_string = link.xpath("preceding::h5[1]/text()")[0].strip()
            row_string = link.text_content().strip()
            row_text = re.sub(r"\s+", " ", row_string)
            docket, name = row_text.split(" - ")
            name, status = name.split("(")

            url = link.xpath(".//a/@href")[0]
            self.cases.append(
                {
                    "docket": docket,
                    "name": name,
                    "status": status[:-1],
                    "date": date_string,
                    "url": f"https://searchdro.kscourts.gov{url}",
                }
            )

        if not self.test_mode_enabled():
            self.close_webdriver_session()



import email
from typing import Optional, TypedDict
from urllib.parse import parse_qs, urlparse

from lxml.html import HtmlElement

from juriscraper.abstract_parser import AbstractParser
from juriscraper.AbstractSite import logger
from juriscraper.lib.email_utils import parse_email_html
from juriscraper.state.texas.common import CourtID

COURT_NAME_TO_ID = {
    "first court of appeals": CourtID.FIRST_COURT_OF_APPEALS.value,
    "second court of appeals": CourtID.SECOND_COURT_OF_APPEALS.value,
    "third court of appeals": CourtID.THIRD_COURT_OF_APPEALS.value,
    "fourth court of appeals": CourtID.FOURTH_COURT_OF_APPEALS.value,
    "fifth court of appeals": CourtID.FIFTH_COURT_OF_APPEALS.value,
    "sixth court of appeals": CourtID.SIXTH_COURT_OF_APPEALS.value,
    "seventh court of appeals": CourtID.SEVENTH_COURT_OF_APPEALS.value,
    "eighth court of appeals": CourtID.EIGHTH_COURT_OF_APPEALS.value,
    "ninth court of appeals": CourtID.NINTH_COURT_OF_APPEALS.value,
    "tenth court of appeals": CourtID.TENTH_COURT_OF_APPEALS.value,
    "eleventh court of appeals": CourtID.ELEVENTH_COURT_OF_APPEALS.value,
    "twelfth court of appeals": CourtID.TWELFTH_COURT_OF_APPEALS.value,
    "thirteenth court of appeals": CourtID.THIRTEENTH_COURT_OF_APPEALS.value,
    "fourteenth court of appeals": CourtID.FOURTEENTH_COURT_OF_APPEALS.value,
    "fifteenth court of appeals": CourtID.FIFTEENTH_COURT_OF_APPEALS.value,
    "court of criminal appeals": CourtID.COURT_OF_CRIMINAL_APPEALS.value,
    "supreme court": CourtID.SUPREME_COURT.value,
}

SUBJECT_PREFIX = "Automated Case Update from "


class TamesEmailData(TypedDict):
    court_id: str
    case_number: str
    url: str


class TamesEmail(AbstractParser):
    """Parse TAMES case notification emails from Texas courts."""

    def __init__(self, court_id: str = "") -> None:
        self.court_id: str = court_id
        self.tree: Optional[HtmlElement] = None
        self.message: Optional[email.message.Message] = None
        self._court_id: Optional[str] = None

    @property
    def data(self) -> Optional[TamesEmailData]:
        if self._court_id is None or self.tree is None:
            raise ValueError("Unable to parse email")

        url = self._parse_case_url()
        if url is None:
            raise ValueError("No url found in email")

        case_number = parse_qs(urlparse(url).query).get("cn", [None])[0]
        if case_number is None:
            logger.error("Unable to extract case number from URL: %s", url)
            return None

        return TamesEmailData(
            court_id=self._court_id,
            url=url,
            case_number=case_number,
        )

    def _parse_text(self, text: str) -> None:
        self.message = email.message_from_string(text)
        subject = self.message.get("Subject", failobj="")

        if not subject.startswith(SUBJECT_PREFIX):
            logger.error(
                "Email subject did not match expected prefix. Subject: %s",
                subject,
            )
            return

        court_name = subject.removeprefix(SUBJECT_PREFIX).lower()
        self._court_id = COURT_NAME_TO_ID.get(court_name)
        if self._court_id is None:
            logger.error("Unknown court name in subject: %s", court_name)
            return

        self.tree = parse_email_html(text)

    def _parse_case_url(self) -> Optional[str]:
        anchor = self.tree.find(".//a")
        if anchor is None:
            logger.error("Unable to find link in email body")
            return None
        return anchor.get("href")



from typing import Union

from juriscraper.lib.string_utils import clean_string
from juriscraper.state.texas.common import (
    CourtID,
    CourtType,
    TexasAppealsCourt,
    TexasAppellateBrief,
    TexasCaseEvent,
    TexasCommonData,
    TexasCommonScraper,
    _parse_appeals_court,
)


class TexasSupremeCourtCaseEvent(TexasCaseEvent):
    """
    Extension of `TexasCaseEvent` to allow for the "Remarks" column in Texas
    Supreme Court dockets.

    :ivar remarks: The text content of the relevant cell in the "Remarks"
    column.
    """

    remarks: str


class TexasSupremeCourtAppellateBrief(TexasAppellateBrief):
    """
    Extension of `TexasAppellateBrief` to allow for the "Remarks" column in
    Texas Supreme Court dockets.

    :ivar remarks: The text content of the relevant cell in the "Remarks"
    column.
    """

    remarks: str


class TexasSupremeCourtDocket(TexasCommonData):
    """
    Extension of the `TexasCommonData` schema with data specific to Texas
    Supreme Court dockets.

    :ivar appeals_court: Information about the appeals court which heard this
    case.
    :ivar case_events: A list of `TexasSupremeCourtCaseEvent` objects
    representing the case events.
    :ivar appellate_briefs: A list of `TexasSupremeCourtAppellateBrief` objects
    representing the appellate briefs.
    """

    appeals_court: TexasAppealsCourt
    case_events: list[TexasSupremeCourtCaseEvent]
    appellate_briefs: list[TexasSupremeCourtAppellateBrief]


class TexasSupremeCourtScraper(TexasCommonScraper):
    """
    Extends the `TexasCommonScraper` class to extract data specific to Texas
    Supreme Court dockets. Unique data extracted is:

    - Appeals court information.
    - Remarks on case events and appellate briefs.
    """

    def __init__(self, court_id: str = "texas_sc"):
        super().__init__(court_id)

    @property
    def data(self) -> Union[TexasSupremeCourtDocket, dict[str, None]]:
        """
        Extract parsed data from an HTML tree. This property returns a
        `TexasSupremeCourtDocket`.

        :return: Parsed data.
        """

        common_data = super().data
        if not common_data:
            return {}

        case_events = [
            TexasSupremeCourtCaseEvent(
                remarks=clean_string(remarks_element.text_content()),
                **event_data,
            )
            for remarks_element, event_data in zip(
                self.events["Remarks"], common_data["case_events"]
            )
        ]
        appellate_briefs = [
            TexasSupremeCourtAppellateBrief(
                remarks=clean_string(remarks_element.text_content()),
                **event_data,
            )
            for remarks_element, event_data in zip(
                self.briefs["Remarks"], common_data["appellate_briefs"]
            )
        ]

        return TexasSupremeCourtDocket(
            court_id=CourtID.SUPREME_COURT.value,
            court_type=CourtType.SUPREME.value,
            appeals_court=_parse_appeals_court(self.tree),
            case_events=case_events,
            appellate_briefs=appellate_briefs,
            docket_number=common_data["docket_number"],
            date_filed=common_data["date_filed"],
            case_type=common_data["case_type"],
            parties=common_data["parties"],
            originating_court=common_data["originating_court"],
            case_name=common_data["case_name"],
            case_name_full=common_data["case_name_full"],
        )



import random
import ssl
import time
from datetime import datetime
from typing import Any

from requests import Session
from requests.adapters import HTTPAdapter
from urllib3.util import create_urllib3_context

from juriscraper.lib.log_tools import make_default_logger

logger = make_default_logger()


class SSLAdapter(HTTPAdapter):
    def __init__(
        self, ssl_version=ssl.PROTOCOL_TLSv1_2, ciphers=None, **kwargs
    ):
        self.ssl_version = ssl_version or ssl.PROTOCOL_TLS
        self.ssl_context = create_urllib3_context(
            ssl_version=self.ssl_version, ciphers=ciphers
        )
        super().__init__(**kwargs)

    def init_poolmanager(self, *args, **kwargs):
        kwargs["ssl_context"] = self.ssl_context
        return super().init_poolmanager(*args, **kwargs)


def add_delay(delay=0, deviation=0):
    """Create a semi-random delay.

    Delay is the number of seconds your program will be stopped for, and
    deviation is the number of seconds that the delay can vary.
    """
    from juriscraper.AbstractSite import logger

    duration = random.randrange(delay - deviation, delay + deviation)
    logger.info(f"Adding a delay of {duration} seconds. Please wait.")
    time.sleep(duration)


class AcmsApiClient:
    """
    Client for interacting with the court's API, responsible for making requests
    and mapping raw responses to a standardized format.
    """

    BASE_URL_TEMPLATE = (
        "https://{court_id}-showdocservices.azurewebsites.us/api/"
    )
    CASE_DETAILS_ENDPOINT = "CaseDetailsByCaseId/{case_id}"
    DOCKET_ENTRIES_ENDPOINT = "DocketEntriesByCaseId/{case_id}"
    ATTACHMENTS_ENDPOINT = "DocketEntryDocumentsByDocketId/{entry_id}"
    MERGE_PDF_ENDPOINT = "MergePDFFiles/"
    GET_PDF_ENDPOINT = "GetMergedFile/"

    def __init__(self, session: Session, court_id: str):
        self.session = session
        self.court_id = court_id
        self.base_url = self.BASE_URL_TEMPLATE.format(court_id=self.court_id)

    def _map_case_details_response(
        self, raw_data: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Maps the raw API response for case details to a standardized dictionary
        format.

        This method extracts and renames specific fields relevant to existing
        logic for parties and metadata. Not all fields from the raw response are
        mapped.

        :param raw_data: The raw dictionary response from the case details API.
        :return: A dictionary containing the mapped case details.
        """
        case_details = {
            "caseId": raw_data.get("pcx_caseid"),
            "caseNumber": raw_data.get("pcx_casenumber"),
            "originatingCaseNumber": raw_data.get("pcx_originatingcasenumber"),
            "caseFormType": raw_data.get("caseFormType"),
            "name": raw_data.get("pcx_name"),
            "caseOpened": raw_data.get("acms_caseopened"),
            "aNumber": raw_data.get("acms_anumber"),
            "receivedDate": raw_data.get("pcx_receiveddate"),
            "decisionDate": raw_data.get("acms_decisiondate"),
            "partyAttorneyList": raw_data.get("acms_partyattorneylist"),
            "shortCaption": raw_data.get("acms_raw_html"),
            "caseType": (
                raw_data["pcx_casetype"].get("pcx_casetypename")
                if raw_data.get("pcx_casetype")
                else None
            ),
            "caseSubType": (
                raw_data["pcx_casesubtype"].get("pcx_casesubtypename")
                if raw_data.get("pcx_casesubtype")
                else None
            ),
            "caseSubSubType": (
                raw_data["pcx_casesubsubtype"].get("pcx_casesubsubtypename")
                if raw_data.get("pcx_casesubsubtype")
                else None
            ),
            "districtCourtName": (
                raw_data["acms_DistrictCourt"].get("pcx_districtname")
                if raw_data.get("acms_DistrictCourt")
                else None
            ),
            "feeStatus": (
                raw_data["acms_feestatusid"].get("acms_feestatusname")
                if raw_data.get("acms_feestatusid")
                else None
            ),
            "byteCount": (
                int(raw_data.get("byteCount"))
                if raw_data.get("byteCount")
                else None
            ),
            "natureOfSuit": (
                raw_data["acms_NatureofSuit"].get("acms_name")
                if raw_data.get("acms_NatureofSuit")
                else None
            ),
        }

        if raw_data.get("pcx_court"):
            case_details["court"] = {
                "name": raw_data["pcx_court"].get("pcx_name"),
                "abbreviatedName": raw_data["pcx_court"].get(
                    "acms_abbreviatedname"
                ),
            }

        return case_details

    def _map_docket_entries_response(
        self, raw_data: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Maps the raw API response for docket entry details to a standardized
        format.

        :param raw_data: The raw dictionary response from the docket entries API
        :return: A dictionary containing the mapped data for entries.
        """
        docket_info = {
            "isCaseSealed": raw_data.get("isCaseSealed"),
            "isUserCaseParticipant": raw_data.get("isUserCaseParticipant"),
            "byteCount": (
                int(raw_data.get("byteCount"))
                if raw_data.get("byteCount")
                else None
            ),
        }

        entries = []
        for row in raw_data.get("matchingDocketEntries", []):
            entry_data = {
                "endDate": row.get("pcx_enddate"),
                "endDateFormatted": datetime.strptime(
                    row.get("pcx_enddate"), "%Y-%m-%d"
                ).strftime("%m/%d/%Y"),
                "entryNumber": row.get("pcx_entrynumber"),
                "docketEntryText": row.get("pcx_eventdescription"),
                "docketEntryId": row.get("pcx_eventid"),
                "createdOn": row.get("createdon"),
                "documentCount": row.get("documentCount"),
                "pageCount": row.get("pageCount"),
                "fileSize": row.get("fileSize"),
                "restrictedPartyFilingDocketEntry": row.get(
                    "restrictedPartyFilingDocketEntry"
                ),
                "restrictedDocsAvailable": row.get("restrictedDocsAvailable"),
            }
            entries.append(entry_data)

        docket_info["docketEntries"] = entries
        return docket_info

    def _map_docket_entry_document(
        self, raw_doc_data: dict[str, Any]
    ) -> dict[str, Any]:
        """
        Maps a raw dictionary representing a docket entry document to a
        standardized format.


        :param raw_doc_data: A single dictionary object for a document.

        :return: A dictionary containing the mapped document details.
                Returns an empty dictionary if raw_doc_data is None or empty.
        """
        if not raw_doc_data:
            return {}

        return {
            "docketDocumentDetailsId": raw_doc_data.get(
                "acms_docketdocumentdetailsid"
            ),
            "name": raw_doc_data.get("acms_name"),
            "documentUrl": raw_doc_data.get("acms_documenturl"),
            "caseFilingDocumentUrl": raw_doc_data.get(
                "acms_casefilingdocumenturl"
            ),
            "documentPermission": raw_doc_data.get("acms_documentpermission"),
            "pageCount": raw_doc_data.get("acms_pagecount"),
            "fileSize": raw_doc_data.get("acms_filesize"),
            "createdOn": raw_doc_data.get("createdon"),
            "billablePages": raw_doc_data.get("billablePages"),
            "cost": raw_doc_data.get("cost"),
            "documentNumber": raw_doc_data.get("acms_documentnumber"),
            "searchValue": raw_doc_data.get("searchValue"),
            "searchTransaction": raw_doc_data.get("searchTransaction"),
        }

    def get_case_details(self, case_id: str) -> dict[str, Any]:
        """
        Fetches and maps case details for a given case ID.

        :param case_id: The unique identifier of the case.
        :return: The mapped case details.
        """
        path = self.CASE_DETAILS_ENDPOINT.format(case_id=case_id)
        url = f"{self.base_url}{path}"
        logger.info(f"Fetching case details from: {url}")
        response = self.session.get(url)
        response.raise_for_status()
        return self._map_case_details_response(response.json())

    def get_docket_entries(self, case_id: str) -> dict[str, Any]:
        """
        Fetches and maps docket entry details for a given case ID.

        :param case_id: The unique identifier of the case.
        :return: The mapped docket entry details.
        """
        path = self.DOCKET_ENTRIES_ENDPOINT.format(case_id=case_id)
        url = f"{self.base_url}{path}"
        logger.info(f"Fetching docket entries from: {url}")
        response = self.session.post(
            url,
            json={
                "csoId": self.session.acms_user_data["CsoId"],
                "contactType": self.session.acms_user_data["ContactType"],
            },
        )
        response.raise_for_status()
        return self._map_docket_entries_response(response.json())

    def get_attachments(
        self,
        docket_entry_id: str,
        is_case_participant: bool,
        is_restricted_party_filing_entry: bool,
    ) -> list[dict[str, Any]]:
        """
        Fetches and maps attachments for a specific docket entry.

        :param docket_entry_id: The ID of the docket entry to retrieve
                                attachments for.
        :param is_case_participant: Indicates if the user is a participant in
                                    the case.
        :param is_restricted_party_filing_entry: Indicates if the docket entry is a
                                                restricted party filing.
        :return: A list of dictionaries, where each dictionary represents an
                 attachment.
        """
        path = self.ATTACHMENTS_ENDPOINT.format(entry_id=docket_entry_id)
        url = f"{self.base_url}{path}"
        logger.info(f"Fetching attachment from: {url}")
        response = self.session.post(
            url,
            json={
                "isUserCaseParticipant": is_case_participant,
                "restrictedPartyFilingDocketEntry": is_restricted_party_filing_entry,
            },
        )
        response.raise_for_status()
        raw_attachments = response.json()
        if not raw_attachments:
            return []

        return [
            self._map_docket_entry_document(attachment)
            for attachment in raw_attachments
        ]

    def download_pdf(
        self,
        document_id: str,
        page_count: int,
        document_url: str,
        max_attempts: int = 90,
        poll_interval: float = 1.0,
    ):
        """
        Requests generation and download of a merged PDF from the remote API.

        This method sends a request to merge a document, polls the status
        of the operation until it is completed, and then retrieves the PDF file
        from the API.

        :param document_id: identifier of the document in the ACMS system.
        :param page_count: number of pages in the document.
        :param document_url: The URL where the source document can be accessed.
        :param max_attempts: Maximum number of polling attempts before timing out.
        :param poll_interval: Seconds to wait between polling attempts.
        :return: The HTTP response containing the merged PDF file.
        """
        # API endpoints for merge and retrieval
        merge_pdf_path = f"{self.base_url}{self.MERGE_PDF_ENDPOINT}"
        get_pdf_path = f"{self.base_url}{self.GET_PDF_ENDPOINT}"

        # Request body specifying the document to merge
        body = {
            "header": True,
            "mergeScope": "External",
            "pagination": False,
            "docketEntryDocuments": [
                {
                    "acms_docketdocumentdetailsid": document_id,
                    "acms_documenturl": document_url,
                    "acms_pagecount": page_count,
                }
            ],
        }
        # Step 1: Request PDF merge operation
        response = self.session.post(merge_pdf_path, json=body)
        response.raise_for_status()
        data = response.json()

        # Extract URL to check merge job status
        status_url = data["statusQueryGetUri"]

        # Step 2: Poll until the merge job is completed or timeout
        for _ in range(max_attempts):
            response = self.session.get(status_url)
            response.raise_for_status()
            data = response.json()
            status = data["runtimeStatus"].lower()

            if status == "completed":
                pdf_id = data["output"]
                break

            time.sleep(poll_interval)
        else:
            # If loop completes without break
            raise TimeoutError(
                f"Merge job did not complete after {max_attempts} attempts."
            )

        # Step 3: Retrieve the merged PDF by its file ID
        response = self.session.get(f"{get_pdf_path}?fileGuid={pdf_id}")
        response.raise_for_status()
        return response



"""Scraper for the Arizona Supreme Court
CourtID: ariz
Court Short Name: Ariz.
Court Contact: aaborns@courts.az.gov

History:
    2013-04-05: Created by Michael Lissner
    2014-02-07: Revised by Taliah Mirmalek
    2026-01-05: Rewritten to use new API by Luis Manzur
"""

import re
from datetime import date, datetime, timedelta
from typing import Optional
from urllib.parse import quote, urlencode, urljoin

from juriscraper.AbstractSite import logger
from juriscraper.lib.html_utils import get_visible_text
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    base_url = "https://www.azcourts.gov"
    court_param = "Supreme"
    search_page_path = "/opinions/SearchOpinionsMemoDecs"
    first_opinion_date = datetime(1998, 1, 1)
    days_interval = 30

    # XPaths for extracting authentication values
    module_xp = "//div[@id='dnn_ContentPane']/div/a/@name"
    tab_xp = "//script[contains(text(), 'g_dnnsfState')]/text()"
    token_xp = "//input[@name='__RequestVerificationToken']/@value"

    # Static API parameters
    params = {
        "searchPhrase": "",
        "decisionType": "",
        "decisionInvolvementType": "",
        "judgeId": "",
        "caseType": "",
        "caseSubType": "",
        "constitutionalityOpinionsOnly": "false",
    }

    # Map API decision types to precedential status
    # Note: "Decord" (Decision Order) entries are procedural orders, not opinions
    status_map = {
        "Opinion": "Published",
        "Memo": "Unpublished",
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.end_date = date.today()
        self.start_date = self.end_date - timedelta(days=7)
        self.url = urljoin(self.base_url, self.search_page_path)
        self.make_backscrape_iterable(kwargs)
        self.opinions_url = (
            f"{self.base_url}/API/Azcourts/Opinions/GetOpinions"
        )

    def _build_api_url(
        self,
        start_date: Optional[date],
        end_date: Optional[date],
        page: int = 0,
    ) -> str:
        """Build the API URL with query parameters.

        :param start_date: Start date
        :param end_date: End date
        :param page: Page number (0-indexed)
        :return: Full API URL
        """
        params = {
            **self.params,
            "startDate": start_date.strftime("%Y-%m-%d"),
            "endDate": end_date.strftime("%Y-%m-%d"),
            "court": self.court_param,
            "page": page,
        }
        return f"{self.opinions_url}?{urlencode(params)}"

    def _set_auth_headers(self) -> None:
        """Extract and set authentication headers from HTML page.

        :return: None
        """
        tab_script = self.html.xpath(self.tab_xp)[0]
        tab_id = re.search(r'"tabId":(\d+)', tab_script).group(1)

        self.request["headers"].update(
            {
                "ModuleId": self.html.xpath(self.module_xp)[0],
                "TabId": tab_id,
                "RequestVerificationToken": self.html.xpath(self.token_xp)[0],
            }
        )

    def _download(self, request_dict=None):
        """Override download to add authentication headers.

        First fetches the search page to get the verification token,
        then makes the API request with proper headers.

        :param request_dict: Optional request parameters
        :return: JSON response from API
        """
        # In test mode, use the parent _download method which handles mock requests
        if self.test_mode_enabled():
            return super()._download(request_dict)

        # Fetch search page and extract auth headers
        self.html = super()._download()
        self._set_auth_headers()

        # Build API URL and fetch JSON
        self.url = self._build_api_url(self.start_date, self.end_date)
        return super()._download(request_dict)

    def _process_html(self) -> None:
        """Parse API JSON response into case dictionaries.

        :return: None
        """
        json_response = self.html

        total_count = json_response.get("TotalCount", 0)
        page_index = json_response.get("PageIndex", 0)
        page_size = json_response.get("PageSize", 10)

        logger.info(
            "Processing page %d: %d total results",
            page_index + 1,
            total_count,
        )

        for opinion in json_response.get("Opinions", []):
            decision_type = opinion["DecisionType"]
            # Skip Decord (Decision Order) - these are procedural orders, not opinions
            if decision_type not in self.status_map:
                continue
            file_url = opinion["FileUrl"]

            html_summary = opinion["Summary"]
            summary = ""
            if html_summary:
                summary = get_visible_text(html_summary).strip()
            # Normalize whitespace in summary
            summary = " ".join(summary.split())

            self.cases.append(
                {
                    "name": titlecase(opinion["Title"]),
                    "docket": opinion["CaseNumber"],
                    "date": opinion["FilingDate"],
                    "url": urljoin(
                        self.base_url, quote(file_url, safe="/:@?=&")
                    ),
                    "status": self.status_map.get(decision_type, "Unknown"),
                    "author": self._extract_author(
                        opinion.get("OpinionJudges", [])
                    ),
                    "summary": summary,
                }
            )

        # Handle pagination if not in test mode
        if not self.test_mode_enabled():
            total_pages = (total_count + page_size - 1) // page_size
            if page_index + 1 < total_pages:
                self._fetch_next_page(page_index + 1)

    def _extract_author(self, judges: list) -> str:
        """Extract the author judge from the list of opinion judges.

        :param judges: List of judge dictionaries from API
        :return: Author name or empty string
        """
        author = next(
            (
                p
                for p in judges
                if p.get("DecisionInvolvementType") == "Author"
            ),
            None,
        )
        if not author:
            return ""
        name = f"{author.get('FirstName', '')} {author.get('MiddleName', '')} {author.get('LastName', '')}".strip()
        name = " ".join(name.split())  # Normalize whitespace
        if author.get("Suffix"):
            name = f"{name}, {author['Suffix']}"
        return name

    def _fetch_next_page(self, page: int) -> None:
        """Fetch next page of results.

        Auth headers are already set from the initial download.

        :param page: Page number to fetch from
        :return: None
        """
        self.url = self._build_api_url(self.start_date, self.end_date, page)
        self.html = super()._download()
        self._process_html()

    def _download_backwards(self, dates: tuple[date, date]) -> None:
        """Make custom date range request for backscraping.

        :param dates: (start_date, end_date) tuple
        :return: None
        """
        self.start_date, self.end_date = dates
        logger.info(
            "Backscraping for range %s to %s", self.start_date, self.end_date
        )

        # Reset URL to search page so _download fetches fresh auth tokens
        self.url = urljoin(self.base_url, self.search_page_path)
        self.html = self._download()
        self._process_html()

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract lower court and judge from the scraped text.

        :param scraped_text: The text to extract from.
        :return: A dictionary with the metadata.
        """
        lower_court_pattern = re.compile(
            r"""
            Appeal\s+from\s+the\s+(?P<lower_court>[^\n]+)\n\s*
            The\s+(?:Honorable\s+)?(?P<lower_court_judge>[^,]+),\s+Judge[^\n]*\n\s*
            No\.\s+(?P<lower_court_number>[^\s]+)
            """,
            re.X | re.MULTILINE | re.DOTALL,
        )

        result = {}

        if match := lower_court_pattern.search(scraped_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()

            lower_court_judge = match.group("lower_court_judge").strip()
            lower_court_number = match.group("lower_court_number").strip()

            if lower_court:
                result["Docket"] = {"appeal_from_str": lower_court}
            if lower_court_judge:
                result.setdefault("OriginatingCourtInformation", {})[
                    "assigned_to_str"
                ] = lower_court_judge
            if lower_court_number:
                result.setdefault("OriginatingCourtInformation", {})[
                    "docket_number"
                ] = lower_court_number

        return result



"""Scraper for the Supreme Court of Michigan
CourtID: mich
Court Short Name: Mich.
Contact: sitefeedback@courts.mi.gov
History:
    - 2014-09-21: Updated by Jon Andersen to handle some fields being missing
    - 2014-08-05: Updated to have a dynamic URL, an oversight during check in.
    - 2022-01-28: Updated for new web site, @satsuki-chan.
    - 2026-02-25: Added backscraping support going back to December 2000.
"""

import re
from datetime import date, datetime
from urllib.parse import urlencode, urljoin

from juriscraper.AbstractSite import logger
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    title_re = re.compile(
        r"(MSC|COA) (?P<docket>\d{6})\s+(?P<name>.+)\s+Opinion"
    )
    court = "Supreme Court"
    days_interval = 10
    first_opinion_date = datetime(2000, 12, 12)
    filters = (
        ("releaseDate", "Past Year"),
        ("page", "1"),
        ("sortOrder", "Newest"),
        ("searchQuery", ""),
        ("resultType", "opinions"),
        ("pageSize", "100"),
    )

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        params = self.filters + (
            ("resource", "supreme-court-opinion"),
            ("aAppellateCourt", self.court),
        )
        self.url = f"https://www.courts.michigan.gov/api/CaseSearch/SearchCaseOpinions?{urlencode(params)}"
        # Currently blocked from Michigan so we drop the user agent for now
        self.request["headers"] = {"User-Agent": ""}
        self.make_backscrape_iterable(kwargs)

    def _process_html(self) -> None:
        """Process the html and extract out the opinions

        :return: None
        """
        for item in self.html["searchItems"]:
            self.cases.append(self._extract_case_data_from_item(item))

    def _extract_case_data_from_item(self, item: dict) -> dict:
        """Extract the case data from the item

        :param item: The item from the API
        :return: The case data
        """
        if match := self.title_re.search(item["title"]):
            docket = match.group("docket")
            name = self.cleanup_case_name(match.group("name"))
        else:
            name, docket = self.get_missing_name_and_docket(item)

        disposition = self.get_disposition(item)

        status = "Published"
        if "Unpublished" in item["title"]:
            status = "Unpublished"

        per_curiam = False
        if "per curiam" in item["title"].lower():
            per_curiam = True

        return {
            "date": item["filingDate"],
            "docket": docket,
            "name": name,
            "url": urljoin(self.url, item["documentUrl"].strip()),
            "lower_court": self.parse_lower_courts(item["courts"]),
            "status": status,
            "per_curiam": per_curiam,
            "disposition": disposition,
        }

    def cleanup_case_name(self, name_raw: str) -> str:
        """Clean up case name in Michigan

        :param name_raw: Raw title string
        :return: Cleaned name
        """
        name = titlecase(re.sub(r"\s+", " ", name_raw.strip()))
        return name.replace("People of Mi ", "People of Michigan ")

    def parse_lower_courts(self, courts: list[str]) -> str:
        """Get the lower courts

        :param courts: Court names as an array
        :return: The lower courts combined into a string
        """
        lower_courts = re.sub(r"\s+", " ", ", ".join(courts).lstrip(", "))
        return titlecase(lower_courts)

    def get_missing_name_and_docket(self, item: dict) -> tuple[str, str]:
        """
        Return 2 placeholders. We can get the values via `extract_from_text`
        To be overriden in `michctapp`
        Sometimes the API returns bad "title" values. See #1648
        """
        return "Placeholder case name", "Placeholder docket number"

    def get_disposition(self, item: dict) -> str:
        """To be overriden by michctapp"""
        return ""

    def _build_backscrape_url(self, dates: tuple[date, date]) -> str:
        """Build URL for backscraping with a custom date range.

        :param dates: (start_date, end_date) tuple
        :return: URL string
        """
        from_date = dates[0].strftime("%m/%d/%Y")
        to_date = dates[1].strftime("%m/%d/%Y")
        params = (
            ("releaseDate", f"Custom Range:{from_date}:{to_date}"),
            ("page", "1"),
            ("sortOrder", "Newest"),
            ("searchQuery", ""),
            ("resultType", "opinions"),
            ("pageSize", "100"),
            ("aAppellateCourt", self.court),
        )
        return f"https://www.courts.michigan.gov/api/CaseSearch/SearchCaseOpinions?{urlencode(params)}"

    def _download_backwards(self, dates: tuple[date, date]) -> None:
        """Download opinions for a specific date range.

        :param dates: (start_date, end_date) tuple from make_backscrape_iterable
        :return: None
        """
        logger.info("Backscraping for range %s %s", *dates)
        self.url = self._build_backscrape_url(dates)
        self.html = self._download()
        self._process_html()

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extracts case names and docket numbers from the document's text

        :param scraped_text: the document's text
        :return: a dictionary in the shape of Courtlistener's models
        """
        pattern = re.compile(
            r"(?P<case_name>([\S ]+\n)+)\n+\s+Docket Nos?. (?P<docket_number>(\d{6}(,? and |, |.)?)+)"
        )
        if match := pattern.search(scraped_text[:1500]):
            docket_number = match.group("docket_number").strip(" .")
            name = self.cleanup_case_name(
                match.group("case_name").replace("\n", ";")
            ).strip("; ")
            return {
                "OpinionCluster": {"case_name": name},
                "Docket": {"case_name": name, "docket_number": docket_number},
            }
        else:
            logger.error(
                "mich: extract_from_text failed",
                extra={"opinion_text": scraped_text[:1024].strip()},
            )
            return {}



import re
from datetime import date, datetime, timedelta
from typing import Optional
from urllib.parse import urlencode, urljoin

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    days_interval = 15
    first_opinion_date = datetime(1995, 6, 1).date()

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.base_url = "https://www.wicourts.gov/supreme/scopin.jsp"
        self.status = "Published"
        self.set_url()
        self.cite_regex = r"20\d{2}\sWI\s\d+"
        self.make_backscrape_iterable(kwargs)

    def set_url(
        self, start: Optional[date] = None, end: Optional[date] = None
    ) -> None:
        """Sets URL with appropriate query parameters

        :param start: start date
        :param end: end date
        :return None
        """
        if not start:
            start = datetime.today() - timedelta(days=15)
            end = datetime.today()

        start = start.strftime("%m-%d-%Y")
        end = end.strftime("%m-%d-%Y")

        params = {
            "range": "None",
            "begin_date": start,
            "end_date": end,
            "sortBy": "date",
            "Submit": "Search",
        }
        self.url = f"{self.base_url}?{urlencode(params)}"

    def _process_html(self) -> None:
        """Process the HTML from wisconsin

        :return: None
        """
        for row in self.html.xpath(".//table/tbody/tr"):
            date, docket, caption, link = row.xpath("./td")
            self.cases.append(
                {
                    "date": date.text,
                    "name": caption.text,
                    "url": urljoin(
                        "https://www.wicourts.gov",
                        link.xpath("./input")[0].name,
                    ),
                    "docket": docket.text,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract citation from text

        :param scraped_text: Text of scraped content
        :return: date filed
        """
        first_line = scraped_text[:100].splitlines()[0]
        if match := re.search(self.cite_regex, first_line):
            return {"Citation": match.group(0)}

        return {}

    def _download_backwards(self, dates: tuple[date, date]) -> None:
        """Set date range from backscraping args and scrape

        :param dates: (start_date, end_date) tuple
        :return None
        """
        logger.info("Backscraping for range %s %s", *dates)
        self.set_url(*dates)
        self.html = self._download()
        self._process_html()



"""Scraper for Louisiana Court of Appeal, First Circuit
CourtID: lactapp_1
Court Short Name: La. Ct. App. 1st Cir.
Author: mmantel
History:
  2019-11-24: Created by mmantel
  2025-07-21: Updated by quevon24
"""

import json
import re
from datetime import datetime

from juriscraper.AbstractSite import logger
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.status = "Unknown"
        self.base_url = "https://lafcca.hosted2.civiclive.com/portal/svc/ContentItemSvc.asmx/GetItemList"
        self.url = self.base_url
        self.method = "POST"
        self.should_have_results = True

        if datetime.now().year != 2026:
            logger.error("Update lactapp_1 parameters from 2026 parameters")

        # This payload is specific for 2026
        self.payload = {
            "parentId": "310729",
            "Params": json.dumps(
                {
                    "ContextId": 310728,
                    "OneLink": "/cms/One.aspx",
                    "RawUrl": "/decisions/o_p_i_n_i_o_n_s_2026",
                    "Extension": "24725",
                    "PortalId": "161585",
                    "PageId": "310727",
                    "InstanceId": "23741",
                }
            ),
        }

        self.parameters = json.dumps(self.payload)

        self.request["headers"] = {
            "Content-Type": "application/json",
            "Accept": "application/json",
            "User-Agent": "Juriscraper",
        }

    def _download(self, request_dict=None):
        if self.test_mode_enabled():
            return super()._download(request_dict)

        self._request_url_post(self.url)
        self._post_process_response()
        return self._return_response_text_object()

    @staticmethod
    def extract_case_name_info(case_name):
        """Extracts docket number and cleans case name

        :param case_name: raw case name with docket number
        :return: (docket_number, clean_name)
        """
        # Match pattern: YYYY XX #### (4 digits year, 2 letters, 4 digits)
        # e.g. 2025 KW 0712, 2025 KM 0194, 2025 KA 0031, 2025 CW 0606, 2025 CJ 0045, 2025 CA 0022
        match = re.match(r"^(\d{4}\s[A-Z]{2}\s\d{4})\s+(.*)", case_name)
        if match:
            return match.group(1), match.group(2).strip()
        return None, case_name  # Return original name if no docket found

    def _process_html(self):
        json_data = self.html
        for item in json_data["d"]["DataObject"]:
            docket, clean_name = self.extract_case_name_info(item["Name"])
            case = {
                "name": clean_name,
                "url": item["DownloadLink"],
                "date": item["CreationDateString"],
                "docket": docket,
                "status": self.status,
            }
            self.cases.append(case)



"""Scraper for the Court of Appeals of Arizona, Division 1
CourtID: arizctapp_div_1
Court Short Name: Ariz. Ct. App.

History:
    2014-02-14: Created by Deb Linton
    2026-01-05: Rewritten to use new API by Luis Manzur
"""

from juriscraper.opinions.united_states.state import ariz


class Site(ariz.Site):
    base_url = "https://coa1.azcourts.gov"
    court_param = "Division1"
    search_page_path = "/Decisions/Search-Decisions"
    module_xp = "//*[@id='dnn_ContentPane']/div[1]/a/@name"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__



# Scraper for Georgia Appeals Court
# CourtID: gactapp
# Court Short Name: gactapp
# Author: Andrei Chelaru
# Reviewer: mlr
# Date created: 25 July 2014

from datetime import date, datetime, timedelta

from dateutil.relativedelta import relativedelta

from juriscraper.AbstractSite import logger
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    first_opinion_date = date(2012, 3, 30)
    days_interval = 7

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        today = date.today()
        last_week = today - timedelta(days=7)
        self.status = "Published"

        # default most recent week
        self.url = (
            "https://www.gaappeals.gov/wp-content/themes/benjamin/docket/"
            f"docketdate/results_all.php?OPstartDate={last_week:%Y-%m-%d}&OPendDate={today:%Y-%m-%d}&submit=Start+Opinions+Search"
        )

        # build backscrape iterable
        self.make_backscrape_iterable(kwargs)

    def _process_html(self):
        for row in self.html.xpath("//tr")[::-1][:-1]:
            docket, name, date_el, disposition, _, url_el = row.xpath(".//td")
            url = url_el.xpath(".//a")[0].get("href")
            if url == "https://efast.gaappeals.gov/download?filingId=":
                logger.error("Skipping doc with empty filingId: %s", url)
                continue
            self.cases.append(
                {
                    "docket": docket.text_content().strip(),
                    "name": titlecase(name.text_content().strip()),
                    "date": date_el.text_content().strip(),
                    "disposition": disposition.text_content().strip().title(),
                    "url": url,
                }
            )

    def _download_backwards(self, date_range: str) -> None:
        """Backscrape one weekly date range (start,end)."""
        start, end = date_range.split(",")
        self.url = (
            "https://www.gaappeals.gov/wp-content/themes/benjamin/docket/"
            f"docketdate/results_all.php?OPstartDate={start}&OPendDate={end}&submit=Start+Opinions+Search"
        )

        self.html = self._download()
        self._process_html()

    def make_backscrape_iterable(self, kwargs: dict[str, str]) -> None:
        """Generate weekly date ranges from backscrape_start → backscrape_end."""
        start_str = kwargs.get("backscrape_start")
        end_str = kwargs.get("backscrape_end")

        if not start_str or not end_str:
            # Default: scrape past 6 months
            end_date = date.today()
            start_date = end_date - relativedelta(months=6)
        else:
            start_date = datetime.strptime(start_str, "%Y-%m-%d").date()
            end_date = datetime.strptime(end_str, "%Y-%m-%d").date()

        date_ranges = []
        current = start_date
        while current <= end_date:
            week_end = min(current + timedelta(days=7), end_date)
            date_ranges.append(f"{current:%Y-%m-%d},{week_end:%Y-%m-%d}")
            current = week_end + timedelta(days=1)

        self.back_scrape_iterable = date_ranges



"""
Author: William Palin
Date created: 2023-01-16
Scraper for the Decisiones del Tribunal Supremo
CourtID: pr
Court Short Name: Puerto Rico
"""

import re
from datetime import date, datetime

from dateparser import parse

from juriscraper.AbstractSite import logger
from juriscraper.lib.string_utils import titlecase
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    first_opinion_date = "1998/01/01"
    base_url = "https://poderjudicial.pr/index.php/tribunal-supremo/decisiones-del-tribunal-supremo/decisiones-del-tribunal-supremo"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.year = date.today().year
        self.url = f"{self.base_url}-{self.year}/"
        self.status = "Published"
        self.request["verify"] = False
        self.make_backscrape_iterable(kwargs)

    def _process_html(self) -> None:
        """Process the HTML and extract opinions into self.cases

        :return: None
        """
        # The website has used 3 different HTML layouts over the years.
        # We detect the format by counting sibling TDs of the TSPR link:
        #   >=5 → middle format (2005-2022)
        #   4   → oldest format (1998-2004)
        #   0   → newest format (2023+)
        #   1-3 → false positive (non-citation link containing "TSPR")
        cells_in_tr_xpath = (
            "ancestor::tr[1]/following-sibling::tr[position() <= 5]"
        )
        cells_in_td_xpath = "ancestor::td[1]/following-sibling::td"

        for link in self.html.xpath("//a[contains(string(.), 'TSPR')]"):
            url = link.xpath("@href")[0]
            citation = link.xpath("string(.)").strip()
            sibling_td_count = len(link.xpath(cells_in_td_xpath))

            if sibling_td_count >= 5:
                # Middle format (2005-2022): single row per case
                # TDs: [citation, materia, docket, name, date, ponente]
                # >= 5 instead of == 5 because some rows on the source
                # merge two cases into a single <tr> (e.g. 2013 TSPR 87)
                cells = link.xpath(cells_in_td_xpath)
                docket = cells[1].text_content().strip()
                name = cells[2].text_content().strip()
                date_str = cells[3].text_content().strip()
            elif sibling_td_count == 4:
                # Oldest format (1998-2004): single row per case
                # TDs: [citation, docket, name, day_or_date, type]
                # The date column varies: sometimes just a day number
                # (with month from a preceding <h3>), sometimes a
                # day-month string like "21-Nov", "7/AGOSTO",
                # "26 ABRIL", or even a full date
                cells = link.xpath(cells_in_td_xpath)
                docket = cells[0].text_content().strip()
                name = cells[1].text_content().strip()
                day_field = re.sub(
                    r"\s+", " ", cells[2].text_content()
                ).strip()
                if day_field.isdigit():
                    month_h3 = link.xpath(
                        "ancestor::table[1]/preceding::h3[1]"
                    )
                    month_name = (
                        month_h3[0].text_content().strip().lower()
                        if month_h3
                        else ""
                    )
                    date_str = f"{day_field} de {month_name} de {self.year}"
                elif day_field:
                    # Append the year if not already present, so
                    # dateparser can resolve "21-Nov" → "21-Nov 2000"
                    date_str = day_field
                    if str(self.year) not in date_str:
                        date_str = f"{date_str} {self.year}"
                else:
                    date_str = ""
            elif sibling_td_count == 0:
                # Newest format (2023+): vertical layout with labeled rows
                # Rows: [Núm., Partes, Ponente, Fecha, Materia]
                cells = link.xpath(cells_in_tr_xpath)
                docket = self.extract_cell(cells, 0)
                name = self.extract_cell(cells, 1)
                date_str = cells[3].xpath(".//td")[1].text_content().strip()
            else:
                # 1-3 siblings: false positive match — a link in a
                # non-citation column whose text happens to contain
                # "TSPR" (e.g. case name "ADOPCION CITA TSPR Y PRSC")
                continue

            # "1ro" is Spanish for "primero" (first day of month),
            # which dateparser doesn't recognize
            date_str = re.sub(r"\b1ro\b", "1", date_str)
            date_obj = parse(date_str, languages=["es"])

            date_filed_is_approximate = False
            if not date_obj:
                # E.g. 2022TSPR106 has the docket number in the date
                # column instead of a date due to a website data error
                date_filed_is_approximate = True
                if self.cases:
                    fallback = self.cases[-1]["date"]
                else:
                    fallback = f"{self.year}-01-01"
                logger.info(
                    "Could not parse date for %s, using approximate date %s",
                    citation,
                    fallback,
                )
                parsed_date = fallback
            else:
                parsed_date = str(date_obj.date())

            self.cases.append(
                {
                    "name": titlecase(name),
                    "url": url,
                    "citation": citation.strip(),
                    "docket": docket,
                    "date": parsed_date,
                    "date_filed_is_approximate": date_filed_is_approximate,
                }
            )

    def extract_cell(self, cells, index) -> str:
        """Extract the value from a labeled table cell in the newest layout

        Each cell contains a label (e.g. "Núm.", "Partes") followed by the
        actual value. The number of blank lines between them varies by year,
        so blank lines are filtered out before indexing.

        :param cells: List of table row elements
        :param index: Index of the cell to extract the value from
        :return: The cell value with whitespace stripped
        """
        lines = [
            line.strip()
            for line in cells[index].text_content().split("\n")
            if line.strip()
        ]
        return lines[1] if len(lines) > 1 else lines[0]

    def make_backscrape_iterable(self, kwargs) -> None:
        """Build the year range iterable used for back-scraping

        :param kwargs: Keyword arguments optionally containing backscrape_start
            and backscrape_end as "YYYY/MM/DD" strings
        :return: None
        """
        start = kwargs.get("backscrape_start")
        end = kwargs.get("backscrape_end")

        if not start:
            return

        start_date = datetime.strptime(start, "%Y/%m/%d")
        end_date = datetime.strptime(end, "%Y/%m/%d")

        self.back_scrape_iterable = range(start_date.year, end_date.year + 1)

    def _download_backwards(self, year) -> None:
        """Download and parse opinions for a single back-scrape year

        :param year: The year to scrape
        :return: None
        """
        self.year = year
        self.url = f"{self.base_url}-{year}/"
        self.html = self._download()
        self._process_html()



"""
Scraper for Maryland Court of Appeals
CourtID: md
Court Short Name: MD
Author: Andrei Chelaru
Date created: 06/27/2014
Court Support: webmaster@mdcourts.gov, mdlaw.library@mdcourts.gov
"""

from datetime import date

from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    base_url = "https://www.mdcourts.gov/cgi-bin/indexlist.pl?court={}&year={}&order=bydate&submit=Submit"
    court = "coa"
    start_year = 1995
    current_year = date.today().year
    empty_cite_strings = {"slip.op.", "."}
    no_judge_strings = {"Order", "PC Order", "Per Curiam"}

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.url = self.base_url.format(self.court, self.current_year)
        self.court_id = self.__module__
        self.disable_certificate_verification()
        self.status = "Published"
        self.make_backscrape_iterable(kwargs)

    def _process_html(self) -> None:
        """Parse HTML into case dictionaries

        :return None
        """
        for row in self.html.xpath("//table//tr[td and not (.//h2)]"):
            url = row.xpath("td//a[contains(@href,'pdf')]/@href")[0]
            docket = row.xpath("td[1]//text()")[0]
            date_filed, _, other_date = row.xpath("td[3]/font/text()")[
                0
            ].partition(" ")
            name = row.xpath("td[5]/font/text()")[0].split("(")[0].strip()

            per_curiam = False
            judge = row.xpath("td[4]/font/text()")[0].split("()")[0].strip()
            if judge in self.no_judge_strings:
                # Other strings in self.no_judge_strings point to
                # Per Curiam opinions
                per_curiam = judge != "Order"
                judge = ""

            cite = row.xpath("td[2]/font/text()")[0].strip()
            if cite in self.empty_cite_strings:
                cite = ""

            self.cases.append(
                {
                    "name": name,
                    "url": url,
                    "judge": judge,
                    "date": date_filed,
                    "docket": docket,
                    "citation": cite,
                    "per_curiam": per_curiam,
                    "other_date": other_date.strip(),
                }
            )

    def make_backscrape_iterable(self, kwargs: dict) -> None:
        """Checks if backscrape start and end arguments have been passed
        by caller, and parses them accordingly

        :param kwargs: passed when initializing the scraper, may or
            may not contain backscrape controlling arguments
        :return None
        """
        start = kwargs.get("backscrape_start")
        end = kwargs.get("backscrape_end")

        start = int(start) if start else self.start_year
        end = int(end) + 1 if end else self.current_year

        self.back_scrape_iterable = range(start, end)

    def _download_backwards(self, year: int) -> None:
        """Build URL with year input and scrape

        :param year: year to scrape
        :return None
        """
        self.url = self.base_url.format(self.court, year)
        self.html = self._download()
        self._process_html()



# Scraper for New York Appellate Term 1st Dept.
# CourtID: nyappterm_1st
# Court Short Name: NY

import re
from datetime import date
from typing import Any

from juriscraper.lib.string_utils import clean_string
from juriscraper.opinions.united_states.state import ny


class Site(ny.Site):
    first_opinion_date = date(2003, 9, 25)

    # If more than 500 results are found, no results will be shown
    days_interval = 30

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court = "Appellate Term, 1st Dept"
        self.court_id = self.__module__
        self._set_parameters()
        self.make_backscrape_iterable(kwargs)

    def extract_from_text(self, scraped_text: str) -> dict[str, Any]:
        """Can we extract the docket number from the text?

        Examples
        2017-1270 K CR
        570086/18
        20-159

        :param scraped_text: The content of the document downloaded
        :return: Metadata to be added to the case
        """
        regexes = [
            re.compile(
                r"<br>[\s\n]*(?P<docket_number>\d{2,}-\d{2,}( [A-Z]{1,2} CR?)?)"
            ),
            re.compile(r"<br>[\s\n]*(?P<docket_number>\d+/\d+)[\s\n]*<br>"),
        ]
        for regex in regexes:
            if match := regex.search(clean_string(scraped_text[:2500])):
                cleaned_docket = self.clean_docket_match(match)
                if cleaned_docket:
                    return {"Docket": {"docket_number": cleaned_docket}}

        return {}



# Author: Michael Lissner
# Date created: 2013-06-03
# Date updated: 2020-02-25

import re
from typing import Union

from juriscraper.AbstractSite import logger
from juriscraper.lib.exceptions import InvalidDocumentError
from juriscraper.OpinionSiteLinear import OpinionSiteLinear


class Site(OpinionSiteLinear):
    # Home: https://juddocumentservice.mt.gov/getDailyOrders
    base_url = "https://juddocumentservice.mt.gov"
    download_base = f"{base_url}/getDocByCTrackId?DocId="
    cite_regex = r"((19|20)\d{2}\sMT\s\d{1,3}[A-Z]?)"

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.court_id = self.__module__
        self.url = f"{self.base_url}/getDailyOrders"
        self.expected_content_types = None
        self.should_have_results = True

    def _process_html(self):
        for row in self.html:
            description = row["documentDescription"]
            if not description.startswith("Opinion"):
                # skip orders and just do opinions
                logger.info("Skipping row with description %s", description)
                continue

            status = (
                "Published" if "Published" in description else "Unpublished"
            )

            docket = row["caseNumber"]
            if docket.startswith("DA"):
                nature = "Direct Appeal"
            elif docket.startswith("OP"):
                nature = "Original Proceeding"
            elif docket.startswith("PR"):
                nature = "Professional Regulation"
            elif docket.startswith("AF"):
                nature = "Administrative File"
            else:
                nature = "Unknown"

            author = ""
            disposition = ""
            summary = ""
            if author_match := re.search(
                r"Justice (?P<author>.*?)\s*(?:author(ed)?|,|-|\.)",
                description,
                re.I,
            ):
                author = author_match.group("author")
                disposition = description[author_match.end() :].strip(" .,-")
                disposition = disposition[:1].upper() + disposition[1:]
            else:
                summary = description

            self.cases.append(
                {
                    "url": self.download_base + row["cTrackId"],
                    "status": status,
                    "date": row["fileDate"],
                    "name": row["title"],
                    "docket": docket,
                    "nature_of_suit": nature,
                    "author": author,
                    "disposition": disposition,
                    "summary": summary,
                }
            )

    def extract_from_text(self, scraped_text: str) -> dict:
        """Extract citation, and OriginatingCourtInformation from text

        :param scraped_text: Text of scraped content
        :return: date filed
        """
        result = {}
        first_text = scraped_text[:400]
        if match := re.search(self.cite_regex, first_text):
            result["Citation"] = match.group(0)

        lower_court_pattern = re.compile(
            r"""
                APPEAL\s+FROM:\s+
                (?P<lower_court>.*?)(?:,\s+Cause)
                (?:.*?No\.\s+(?P<lower_court_docket>[^\n,]+))?
                (?:.*?Honorable\s+(?P<judge>.+?),\s*Presiding\s*Judge)?
            """,
            re.X | re.DOTALL,
        )

        if match := lower_court_pattern.search(scraped_text):
            lower_court = re.sub(
                r"\s+", " ", match.group("lower_court")
            ).strip()
            result.setdefault("Docket", {})["appeal_from_str"] = lower_court
            lower_court_judge = match.group("judge")
            if lower_court_judge:
                result.setdefault("OriginatingCourtInformation", {})[
                    "assigned_to_str"
                ] = lower_court_judge
            lower_court_docket = match.group("lower_court_docket")
            if lower_court_docket:
                result.setdefault("OriginatingCourtInformation", {})[
                    "docket_number"
                ] = lower_court_docket

        docket_number_pattern = re.compile(r"(DA|OP)\s[0-9]{2}-[0-9]{4}")
        if match := docket_number_pattern.search(scraped_text):
            result.setdefault("Docket", {})["docket_number"] = match.group(0)

        return result

    @staticmethod
    def cleanup_content(content: Union[str, bytes]) -> str:
        """Raise an error if the content is invalid; otherwise just return it

        Not cleaning up in the common sense; but avoids ingesting error
        pages. This source does not mark the error page with an error status
        and does not have content type headers; so we can't detect the error
        through standard controls

        The content may be a string if it's a string error response, like in
        https://storage.courtlistener.com/txt/2025/01/17/mercer_v._dphhs.txt

        When the content is PDF bytes, it contains the
        'DOCUMENT UNAVAILABLE FOR VIEWING' string. However, we can't access
        that value directly on the bytes. Luckyly, all those PDFs have the
        same hash, so it won't be a big problem. Check #1323 for more info

        :param content: the downloaded content; maybe a string or PDF bytes
        :return: the downloaded content, unchanged
        """
        if isinstance(content, str) and (
            "No document found with CTrack ID" in content[:1000]
            or "Your support ID is" in content[:1000]
        ):
            raise InvalidDocumentError(content)

        return content
